from canaille.oidc.installation import install


def test_install_jwks(configuration):
    del configuration["CANAILLE_OIDC"]["ACTIVE_JWKS"]

    install(configuration, debug=False)
    assert "ACTIVE_JWKS" not in configuration["CANAILLE_OIDC"]

    install(configuration, debug=True)
    assert configuration["CANAILLE_OIDC"]["ACTIVE_JWKS"]



from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
from joserfc.jwk import OKPKey


def make_default_jwk(seed=None):
    """Generate a deterministic JWK based on a seed if available.

    If seed is provided, it will be used to generate a deterministic Ed25519 key.
    Otherwise, a random Ed25519 key is generated.
    """
    if seed:
        salt = b"canaille-jwk-generation-salt-v1"

        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=salt,
            iterations=100000,
            backend=default_backend(),
        )
        key_material = kdf.derive(seed.encode())

        private_key = ed25519.Ed25519PrivateKey.from_private_bytes(key_material)
    else:
        private_key = ed25519.Ed25519PrivateKey.generate()

    pem = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    key = OKPKey.import_key(pem)
    key.ensure_kid()

    return key



import functools
import sys
from pathlib import Path

import click
from flask import current_app
from flask.cli import with_appcontext

from canaille.backends import Backend

try:
    import hypercorn  # noqa: F401

    HAS_HYPERCORN = True
except ImportError:  # pragma: no cover
    HAS_HYPERCORN = False


def with_backendcontext(func):
    @functools.wraps(func)
    def _func(*args, **kwargs):
        if not current_app.config["TESTING"]:  # pragma: no cover
            with Backend.instance.session():
                result = func(*args, **kwargs)

        else:
            result = func(*args, **kwargs)

        return result

    return _func


@click.command()
@click.pass_context
@with_appcontext
def install(ctx):
    """Installs canaille elements from the configuration.

    For instance, depending on the configuration, this can generate OIDC
    keys or install LDAP schemas.
    """
    from canaille.app.installation import install

    app = ctx.obj.load_app()
    install(app)


if HAS_HYPERCORN:  # pragma: no cover

    @click.command()
    @with_appcontext
    @click.option(
        "--config",
        default=None,
        help="Path to a TOML hypercorn configuration file.",
        type=click.Path(),
    )
    def run(config: str | None):
        """Run Canaille in a hypercorn application server.

        Have a look at the :doc:`Hypercorn configuration documentation <hypercorn:how_to_guides/configuring>` to find how to configure it.
        """
        from hypercorn.config import Config
        from hypercorn.run import run as hypercorn_run

        config_obj = Config.from_toml(config) if config else Config()
        config_obj.application_path = "canaille.app.server:app"
        exitcode = hypercorn_run(config_obj)
        sys.exit(exitcode)


@click.group()
def config():
    """Handle Canaille configuration file."""
    pass


@config.command()
@with_appcontext
@click.option(
    "--path", default=None, type=click.Path(), help="The path to the config file"
)
def dump(path: Path | None):
    """Export the configuration in TOML format.

    The configuration is exported to the file path passed by ``--path`` if set.
    else it is displayed on the standard output.
    """
    from canaille.app.configuration import settings_factory
    from canaille.app.toml import export_config

    config_obj = settings_factory(current_app.config, init_with_examples=True)
    content = export_config(config_obj)
    if not path:
        click.echo(content, nl=False)

    else:
        with open(path, "w") as fd:
            fd.write(content)
        click.echo(f"Wrote configuration file at {path}")


@config.command()
@with_appcontext
@with_backendcontext
def check():
    """Test the network connections defined in the configuration file.

    Attempt to reach the database and the SMTP server with the provided
    credentials.
    """
    from canaille.app.configuration import check_network_config

    success = True
    results = check_network_config(current_app.config)
    prefix = {
        True: click.style("OK", fg="green"),
        False: click.style("KO", fg="red"),
        None: click.style("--", fg="blue"),
    }
    for result in results:
        success = result.success is not False and success
        click.echo(prefix[result.success], nl=False)
        click.echo(" " + result.message)

    if not success:
        sys.exit(1)


def register(cli):
    cli.add_command(install)
    if HAS_HYPERCORN:  # pragma: no branch
        cli.add_command(run)
    cli.add_command(config)



from unittest import mock

from canaille.app import models


def test_user_without_password_first_login(testclient, backend, smtpd):
    assert len(smtpd.messages) == 0
    u = models.User(
        formatted_name="Temp User",
        family_name="Temp",
        user_name="temp",
        emails=["john@doe.test", "johhny@doe.test"],
    )
    backend.save(u)

    res = testclient.get("/login", status=200)
    res.form["login"] = "temp"
    res = res.form.submit(status=302)

    assert res.location == "/firstlogin/temp"
    res = res.follow(status=200)
    res.mustcontain("First login")

    res = res.form.submit(name="action", value="sendmail")
    assert (
        "success",
        "A password initialization link has been sent at your email address. "
        "You should receive it within a few minutes.",
    ) in res.flashes
    assert len(smtpd.messages) == 2
    assert [message["X-RcptTo"] for message in smtpd.messages] == u.emails
    assert "Password initialization" in smtpd.messages[0].get("Subject")
    backend.delete(u)


@mock.patch("smtplib.SMTP")
def test_first_login_account_initialization_mail_sending_failed(
    SMTP, testclient, backend, smtpd
):
    SMTP.side_effect = mock.Mock(side_effect=OSError("unit test mail error"))
    assert len(smtpd.messages) == 0

    u = models.User(
        formatted_name="Temp User",
        family_name="Temp",
        user_name="temp",
        emails=["john@doe.test"],
    )
    backend.save(u)

    res = testclient.get("/firstlogin/temp")
    res = res.form.submit(name="action", value="sendmail", expect_errors=True)
    assert (
        "success",
        "A password initialization link has been sent at your email address. "
        "You should receive it within a few minutes.",
    ) not in res.flashes
    assert ("error", "Could not send the password initialization email") in res.flashes
    assert len(smtpd.messages) == 0
    backend.delete(u)


def test_first_login_form_error(testclient, backend, smtpd):
    assert len(smtpd.messages) == 0
    u = models.User(
        formatted_name="Temp User",
        family_name="Temp",
        user_name="temp",
        emails=["john@doe.test"],
    )
    backend.save(u)

    res = testclient.get("/firstlogin/temp", status=200)
    res.form["csrf_token"] = "invalid"
    res = res.form.submit(
        name="action", value="sendmail", status=400, expect_errors=True
    )
    assert len(smtpd.messages) == 0
    backend.delete(u)


def test_first_login_page_unavailable_for_users_with_password(
    testclient, backend, user
):
    testclient.get("/firstlogin/user", status=404)


def test_user_password_deleted_during_login(testclient, backend):
    u = models.User(
        formatted_name="Temp User",
        family_name="Temp",
        user_name="temp",
        emails=["john@doe.test"],
        password="correct horse battery staple",
    )
    backend.save(u)

    res = testclient.get("/login")
    res.form["login"] = "temp"
    res = res.form.submit().follow()
    res.form["password"] = "correct horse battery staple"

    u.password = None
    backend.save(u)

    res = res.form.submit(status=302)
    assert res.location == "/firstlogin/temp"

    backend.delete(u)


def test_smtp_disabled(testclient, backend, smtpd):
    testclient.app.config["CANAILLE"]["SMTP"] = None

    assert len(smtpd.messages) == 0
    u = models.User(
        formatted_name="Temp User",
        family_name="Temp",
        user_name="temp",
        emails=["john@doe.test", "johhny@doe.test"],
    )
    backend.save(u)

    res = testclient.get("/login", status=200)
    res.form["login"] = "temp"
    res = res.form.submit()
    res = res.follow()
    res.form["password"] = "incorrect horse"
    res = res.form.submit()
    assert ("error", "Login failed, please check your information") in res.flashes
    res.form["password"] = ""
    res = res.form.submit()
    assert ("error", "Login failed, please check your information") in res.flashes

    backend.delete(u)



import datetime
import uuid

from authlib.common.urls import add_params_to_uri
from authlib.oauth2 import OAuth2Error
from authlib.oauth2.rfc6749.errors import InvalidRequestError
from authlib.oidc.core.errors import ConsentRequiredError
from flask import Blueprint
from flask import abort
from flask import current_app
from flask import flash
from flask import g
from flask import jsonify
from flask import redirect
from flask import request
from flask import session
from flask import url_for
from joserfc import jwt
from joserfc.errors import JoseError
from werkzeug.datastructures import CombinedMultiDict
from werkzeug.exceptions import HTTPException

from canaille.app import models
from canaille.app.flask import cache
from canaille.app.flask import csrf
from canaille.app.i18n import gettext as _
from canaille.app.session import logout_user
from canaille.app.templating import render_template
from canaille.backends import Backend
from canaille.core.auth import AuthenticationSession
from canaille.core.auth import redirect_to_next_auth_step

from ..oauth import ClientConfigurationEndpoint
from ..oauth import ClientRegistrationEndpoint
from ..oauth import IntrospectionEndpoint
from ..oauth import RevocationEndpoint
from ..oauth import UserInfoEndpoint
from ..oauth import authorization
from ..oauth import get_issuer
from ..oauth import registry
from ..oauth import server_jwks
from ..utils import SCOPE_DETAILS
from .forms import AuthorizeForm
from .forms import LogoutForm
from .well_known import openid_configuration

bp = Blueprint("endpoints", __name__, url_prefix="/oauth")

AUTHORIZATION_REQUEST_PROCESS_TIMEOUT: int = 300


@bp.errorhandler(HTTPException)
def http_error_handler(error):
    return {
        "error": error.name.lower().replace(" ", "_"),
        "error_description": error.description,
    }, error.code


def save_authorization_request_datetime(request_url: str, now: datetime.datetime):
    key = f"auth-request:{request_url}"
    if not cache.get(key):
        cache.set(key, now.isoformat(), timeout=AUTHORIZATION_REQUEST_PROCESS_TIMEOUT)


def get_authorization_request_datetime(request_url: str) -> datetime.datetime | None:
    key = f"auth-request:{request_url}"
    return datetime.datetime.fromisoformat(cache.get(key)) if cache.get(key) else None


@bp.route("/authorize", methods=["GET", "POST"])
@csrf.exempt
def authorize():
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    current_app.logger.debug(
        "authorization endpoint request:\nGET: %s\nPOST: %s",
        request.args.to_dict(flat=False),
        request.form.to_dict(flat=False),
    )

    redirect_url = (
        request.url
        if request.method == "GET"
        else add_params_to_uri(request.url, request.form)
    )

    try:
        # Check that the request is well-formed
        check_prompt_value()

        # Check that login is needed
        if response := authorize_login(redirect_url, now):
            return response

        # Get the user consent if needed
        return authorize_consent(redirect_url, now)

    except OAuth2Error as error:
        if not error.redirect_uri:
            return render_template(
                "error.html",
                description=error.get_error_description(),
                error_code=error.status_code,
            ), error.status_code

        return authorization.handle_error_response(request, error)


def check_prompt_value():
    # Until this is fixed upstream:
    # https://github.com/lepture/authlib/issues/735
    #
    # https://openid.net/specs/openid-connect-prompt-create-1_0.html#name-authorization-request
    # If the OpenID Provider receives a prompt value that it does
    # not support (not declared in the prompt_values_supported
    # metadata field) the OP SHOULD respond with an HTTP 400 (Bad
    # Request) status code and an error value of invalid_request.
    # It is RECOMMENDED that the OP return an error_description
    # value identifying the invalid parameter value.
    if (
        request.values.get("prompt")
        and request.values["prompt"]
        not in openid_configuration()["prompt_values_supported"]
    ):
        raise InvalidRequestError(
            description=f"prompt '{request.values['prompt']}' value is not supported",
            redirect_uri=request.values.get("redirect_uri"),
        )


def start_oidc_auth_session(client_id, username=None, prompt=None):
    """Create and save an AuthenticationSession for OIDC authentication.

    :param client_id: The OIDC client identifier
    :param username: The username to authenticate
    :param prompt: The OIDC prompt parameter value
    """
    g.auth = AuthenticationSession(
        user_name=username,
        data={
            "prompt_login": prompt == "login",
            "client_id": client_id,
        },
        template="oidc/auth.html",
    )
    g.auth.save()


def authorize_login(redirect_url, now):
    """If user authentication or registration is needed, return a redirection to the correct page."""
    save_authorization_request_datetime(redirect_url, now)

    if user := g.session and g.session.user:
        auth_time = g.session.last_login_datetime
        if request.values.get(
            "prompt"
        ) == "login" and auth_time < get_authorization_request_datetime(redirect_url):
            session["redirect-after-login"] = redirect_url

            start_oidc_auth_session(
                client_id=request.values.get("client_id"),
                username=g.session.user.user_name,
                prompt=request.values.get("prompt"),
            )
            return redirect_to_next_auth_step()

        # hotfix for
        # https://github.com/lepture/authlib/issues/741
        max_age = auth_time + datetime.timedelta(
            seconds=int(request.values.get("max_age", 0))
        )
        if request.values.get("max_age") and max_age < now:
            session["redirect-after-login"] = redirect_url

            start_oidc_auth_session(
                client_id=request.values.get("client_id"),
                username=g.session.user.user_name,
                prompt=request.values.get("prompt"),
            )
            return redirect_to_next_auth_step()

        if not user.can_use_oidc:
            abort(
                403,
                "The user does not have the permission to achieve OIDC authentication.",
            )

    else:
        if request.values.get("prompt") == "create":
            session["redirect-after-login"] = redirect_url
            return redirect(url_for("core.account.join"))

        if request.values.get("prompt") != "none":
            session["redirect-after-login"] = redirect_url

            start_oidc_auth_session(
                client_id=request.values.get("client_id"),
                prompt=request.values.get("prompt"),
            )
            return redirect(url_for("core.auth.login"))


def is_consent_needed(grant, redirect_url) -> bool:
    """Check whether the consent page must be displayed."""
    consents = Backend.instance.query(
        models.Consent,
        client=grant.client,
        subject=g.session.user,
    )
    consent = consents[0] if consents else None

    if consent and consent.revoked:
        return True

    if (
        request.values.get("prompt") == "consent"
        and consent
        and consent.issue_date < get_authorization_request_datetime(redirect_url)
    ):
        return True

    if grant.client.trusted:
        return False

    requested_scopes = request.values.get("scope", "")
    allowed_scopes = grant.client.get_allowed_scope(requested_scopes).split(" ")
    if consent and all(scope in set(consent.scope) for scope in allowed_scopes):
        return False

    return True


def create_or_update_consent(grant, user, now):
    consents = Backend.instance.query(
        models.Consent,
        client=grant.client,
        subject=user,
    )
    consent = consents[0] if consents else None

    requested_scopes = request.values.get("scope", "")
    allowed_scopes = grant.client.get_allowed_scope(requested_scopes).split(" ")

    if consent:
        if consent.revoked:
            consent.restore()
        consent.issue_date = now
        consent.scope = grant.client.get_allowed_scope(
            " ".join(set(allowed_scopes + consents[0].scope))
        ).split(" ")
    else:
        consent = models.Consent(
            consent_id=str(uuid.uuid4()),
            client=grant.client,
            subject=user,
            scope=allowed_scopes,
            issue_date=now,
        )
    Backend.instance.save(consent)
    current_app.logger.security(
        f"New consent for {user.user_name} in client {consent.client.client_name}"
    )


def authorize_consent(redirect_url, now):
    user = g.session and g.session.user
    grant = authorization.get_consent_grant(end_user=user)

    # Get the authorization code, or display the user consent form
    if request.method == "GET" or "answer" not in request.form:
        if not is_consent_needed(grant, redirect_url):
            return authorization.create_authorization_response(
                grant=grant, grant_user=user
            )

        # https://github.com/lepture/authlib/issues/740
        #
        # consent_required
        # The Authorization Server requires End-User consent.
        # This error MAY be returned when the prompt parameter value in the
        # Authentication Request is none, but the Authentication Request cannot
        # be completed without displaying a user interface for End-User consent.
        elif request.values.get("prompt") == "none":
            raise ConsentRequiredError(redirect_uri=request.values["redirect_uri"])

        form = AuthorizeForm(request.form or None)
        form.action = redirect_url
        return render_template(
            "oidc/authorize.html",
            user=user,
            grant=grant,
            client=grant.client,
            menu=False,
            scope_details=SCOPE_DETAILS,
            ignored_scopes=["openid"],
            form=form,
        )

    if request.form["answer"] == "logout":
        session["redirect-after-login"] = redirect_url
        return redirect(url_for("core.auth.logout"))

    if request.form["answer"] == "deny":
        grant_user = None

    if request.form["answer"] == "accept":
        grant_user = user
        create_or_update_consent(grant, grant_user, now)

    response = authorization.create_authorization_response(
        grant=grant, grant_user=grant_user
    )

    current_app.logger.debug("authorization endpoint response: %s", response.location)
    return response


@bp.route("/token", methods=["POST"])
@csrf.exempt
def issue_token():
    request_params = request.form.to_dict(flat=False)
    grant_type = (
        request_params["grant_type"][0] if request_params["grant_type"] else None
    )
    current_app.logger.debug("token endpoint request: POST: %s", request_params)
    response = authorization.create_token_response()
    current_app.logger.debug("token endpoint response: %s", response.json)

    if response.json.get("access_token"):
        access_token = response.json["access_token"]
        token = Backend.instance.get(models.Token, access_token=access_token)
        if token.subject:
            current_app.logger.security(
                f"Issued {grant_type} token for {token.subject.user_name} in client {token.client.client_name}"
            )
        else:
            current_app.logger.security(
                f"Issued {grant_type} token for client {token.client.client_name}"
            )

    return response


@bp.route("/introspect", methods=["POST"])
@csrf.exempt
def introspect_token():
    current_app.logger.debug(
        "introspection endpoint request: POST: %s", request.form.to_dict(flat=False)
    )
    response = authorization.create_endpoint_response(
        IntrospectionEndpoint.ENDPOINT_NAME
    )
    current_app.logger.debug("introspection endpoint response: %s", response.json)
    return response


@bp.route("/revoke", methods=["POST"])
@csrf.exempt
def revoke_token():
    current_app.logger.debug(
        "revokation endpoint request: POST: %s", request.form.to_dict(flat=False)
    )
    response = authorization.create_endpoint_response(RevocationEndpoint.ENDPOINT_NAME)
    current_app.logger.debug("revokation endpoint response: %s", response.json)
    return response


@bp.route("/register", methods=["POST"])
@csrf.exempt
def client_registration():
    current_app.logger.debug(
        "client registration endpoint request: POST: %s",
        request.json,
    )

    response = authorization.create_endpoint_response(
        ClientRegistrationEndpoint.ENDPOINT_NAME
    )
    current_app.logger.debug("client registration endpoint response: %s", response.json)
    return response


@bp.route("/register/<client_id>", methods=["GET", "PUT", "DELETE"])
@csrf.exempt
def client_registration_management(client_id):
    if request.method == "PUT":
        current_app.logger.debug(
            "client registration management endpoint request: PUT: %s",
            request.json,
        )
    response = authorization.create_endpoint_response(
        ClientConfigurationEndpoint.ENDPOINT_NAME
    )
    current_app.logger.debug(
        "client registration management endpoint response: %s", response.json
    )
    return response


@bp.route("/jwks.json")
def jwks():
    jwks = server_jwks()
    return jsonify(jwks.as_dict(private=False))


@bp.route("/userinfo", methods=["GET", "POST"])
@csrf.exempt
def userinfo():
    current_app.logger.debug("userinfo endpoint request: %s", request.args)
    response = authorization.create_endpoint_response(UserInfoEndpoint.ENDPOINT_NAME)
    current_app.logger.debug("userinfo endpoint response: %s", response)
    return response


@bp.route("/end_session", methods=["GET", "POST"])
@csrf.exempt
def end_session():
    data = CombinedMultiDict((request.args, request.form))
    user = g.session and g.session.user

    if not user:
        return redirect(url_for("core.account.index"))

    form = LogoutForm(request.form)
    form.action = url_for("oidc.endpoints.end_session_submit")

    client = None
    valid_uris = []

    if "client_id" in data:
        client = Backend.instance.get(models.Client, client_id=data["client_id"])
        if client and client.post_logout_redirect_uris:
            valid_uris = client.post_logout_redirect_uris

    if (
        not data.get("id_token_hint")
        or (data.get("logout_hint") and data["logout_hint"] != user.user_name)
    ) and not session.get("end_session_confirmation"):
        session["end_session_data"] = data
        return render_template("oidc/logout.html", form=form, client=client, menu=False)

    if data.get("id_token_hint"):
        try:
            id_token = jwt.decode(
                data["id_token_hint"],
                server_jwks(),
                registry=registry,
            )
        except JoseError as exc:
            return jsonify(
                {
                    "status": "error",
                    "message": str(exc),
                }
            )

        if not id_token.claims["iss"] == get_issuer():
            return jsonify(
                {
                    "status": "error",
                    "message": "id_token_hint has not been issued here",
                }
            )

        if "client_id" in data:
            if (
                data["client_id"] != id_token.claims["aud"]
                and data["client_id"] not in id_token.claims["aud"]
            ):
                return jsonify(
                    {
                        "status": "error",
                        "message": "id_token audience and client_id don't match",
                    }
                )

        else:
            client_ids = (
                id_token.claims["aud"]
                if isinstance(id_token.claims["aud"], list)
                else [id_token.claims["aud"]]
            )
            for client_id in client_ids:
                client = Backend.instance.get(models.Client, client_id=client_id)
                if client:
                    valid_uris.extend(client.post_logout_redirect_uris or [])

        if user.user_name != id_token.claims["sub"] and not session.get(
            "end_session_confirmation"
        ):
            session["end_session_data"] = data
            return render_template(
                "oidc/logout.html", form=form, client=client, menu=False
            )

    logout_user()

    if "end_session_confirmation" in session:
        del session["end_session_confirmation"]

    if (url := data.get("post_logout_redirect_uri")) and data[
        "post_logout_redirect_uri"
    ] in valid_uris:
        if "state" in data:
            url = add_params_to_uri(url, dict(state=data["state"]))
        return redirect(url)

    flash(_("You have been disconnected"), "success")
    return redirect(url_for("core.account.index"))


@bp.route("/end_session_confirm", methods=["POST"])
@csrf.exempt
def end_session_submit():
    form = LogoutForm(request.form)
    form.validate()

    data = session["end_session_data"]
    del session["end_session_data"]

    if request.form["answer"] == "logout":
        session["end_session_confirmation"] = True
        url = add_params_to_uri(url_for("oidc.endpoints.end_session"), data)
        return redirect(url)

    flash(_("You have not been disconnected"), "info")

    return redirect(url_for("core.account.index"))



from typing import Annotated
from typing import Any
from typing import ClassVar
from typing import Literal

from flask import url_for
from pydantic import EmailStr
from pydantic import Field
from scim2_models import AnyExtension
from scim2_models import AuthenticationScheme
from scim2_models import Bulk
from scim2_models import CaseExact
from scim2_models import ChangePassword
from scim2_models import ComplexAttribute
from scim2_models import ETag
from scim2_models import Extension
from scim2_models import ExternalReference
from scim2_models import Filter
from scim2_models import Meta
from scim2_models import Mutability
from scim2_models import Patch
from scim2_models import Reference
from scim2_models import Required
from scim2_models import Resource
from scim2_models import ResourceType
from scim2_models import Returned
from scim2_models import Schema
from scim2_models import SchemaExtension
from scim2_models import ServiceProviderConfig
from scim2_models import Sort
from scim2_models import Uniqueness

from canaille.app import DOCUMENTATION_URL


class Name(ComplexAttribute):
    formatted: Annotated[str | None, Required.true] = None
    """The full name, including all middle names, titles, and suffixes as
    appropriate, formatted for display (e.g., 'Ms. Barbara J Jensen, III')."""

    family_name: Annotated[str | None, Required.true] = None
    """The family name of the User, or last name in most Western languages
    (e.g., 'Jensen' given the full name 'Ms. Barbara J Jensen, III')."""

    given_name: str | None = None
    """The given name of the User, or first name in most Western languages
    (e.g., 'Barbara' given the full name 'Ms. Barbara J Jensen, III')."""


class Email(ComplexAttribute):
    value: Annotated[EmailStr | None, Required.true] = None
    """Email addresses for the user."""


class PhoneNumber(ComplexAttribute):
    value: Annotated[str | None, Required.true] = None
    """Phone number of the User."""


class Photo(ComplexAttribute):
    value: Annotated[
        Reference[ExternalReference] | None,
        Mutability.read_only,
        CaseExact.true,
        Required.true,
    ] = None
    """URL of a photo of the User."""


class Address(ComplexAttribute):
    formatted: str | None = None
    """The full mailing address, formatted for display or use with a mailing
    label."""

    street_address: str | None = None
    """The full street address component, which may include house number,
    street name, P.O.

    box, and multi-line extended street address information.
    """

    locality: str | None = None
    """The city or locality component."""

    region: str | None = None
    """The state or region component."""

    postal_code: str | None = None
    """The zip code or postal code component."""

    country: Annotated[str | None, Mutability.read_only, Returned.never] = None
    """The country name component (not used by Canaille)."""

    type: Annotated[str | None, Mutability.read_only, Returned.never] = None
    """A label indicating the attribute's function (not used by Canaille)."""


class GroupMembership(ComplexAttribute):
    value: Annotated[str | None, Mutability.read_only] = None
    """The identifier of the User's group."""

    ref: Annotated[
        Reference[Literal["User"] | Literal["Group"]] | None,
        Mutability.read_only,
    ] = Field(None, serialization_alias="$ref")
    """The reference URI of a target resource, if the attribute is a
    reference."""

    display: Annotated[str | None, Mutability.read_only] = None
    """A human-readable name, primarily used for display purposes."""


class User(Resource[AnyExtension]):
    schemas: Annotated[list[str], Required.true] = [
        "urn:ietf:params:scim:schemas:core:2.0:User"
    ]

    external_id: Annotated[str | None, Mutability.read_only, Returned.never] = None

    user_name: Annotated[
        str | None, Uniqueness.server, Required.true, Mutability.immutable
    ] = None
    """Unique identifier for the User, typically used by the user to directly
    authenticate to the service provider."""

    name: Annotated[Name | None, Required.true] = None
    """The components of the user's real name."""

    Name: ClassVar[type[ComplexAttribute]] = Name

    display_name: str | None = None
    """The name of the User, suitable for display to end-users."""

    profile_url: Reference[ExternalReference] | None = None
    """A fully qualified URL pointing to a page representing the User's online
    profile."""

    title: str | None = None
    """The user's title, such as "Vice President"."""

    preferred_language: str | None = None
    """Indicates the User's preferred written or spoken language.

    Generally used for selecting a localized user interface; e.g.,
    'en_US' specifies the language English and country US.
    """

    active: Annotated[bool | None, Required.true] = None
    """A Boolean value indicating the User's administrative status."""

    password: Annotated[str | None, Mutability.write_only, Returned.never] = None
    """The User's cleartext password."""

    emails: list[Email] | None = None
    """Email addresses for the user."""

    Emails: ClassVar[type[ComplexAttribute]] = Email

    phone_numbers: list[PhoneNumber] | None = None
    """Phone numbers for the User."""

    PhoneNumbers: ClassVar[type[ComplexAttribute]] = PhoneNumber

    photos: Annotated[list[Photo] | None, Mutability.read_only] = None
    """URLs of photos of the User."""

    Photos: ClassVar[type[ComplexAttribute]] = Photo

    addresses: list[Address] | None = None
    """A physical mailing address for this User."""

    Addresses: ClassVar[type[ComplexAttribute]] = Address

    groups: Annotated[list[GroupMembership] | None, Mutability.read_only] = None
    """A list of groups to which the user belongs, either through direct
    membership, through nested groups, or dynamically calculated."""

    Groups: ClassVar[type[ComplexAttribute]] = GroupMembership


class GroupMember(ComplexAttribute):
    value: Annotated[str | None, Mutability.immutable, Required.true] = None
    """Identifier of the member of this Group."""

    ref: Annotated[
        Reference[Literal["User"] | Literal["Group"]] | None,
        Mutability.immutable,
        Required.true,
    ] = Field(None, serialization_alias="$ref")
    """The reference URI of a target resource, if the attribute is a
    reference."""

    display: Annotated[str | None, Mutability.read_only] = None
    """A human-readable name, primarily used for display purposes."""


class Group(Resource[Any]):
    schemas: Annotated[list[str], Required.true] = [
        "urn:ietf:params:scim:schemas:core:2.0:Group"
    ]

    external_id: Annotated[str | None, Mutability.read_only, Returned.never] = None

    display_name: Annotated[str | None, Required.true, Mutability.immutable] = None
    """A human-readable name for the Group."""

    members: Annotated[list[GroupMember] | None, Required.true] = None
    """A list of members of the Group."""

    Members: ClassVar[type[ComplexAttribute]] = GroupMember


class EnterpriseUser(Extension):
    schemas: Annotated[list[str], Required.true] = [
        "urn:ietf:params:scim:schemas:extension:enterprise:2.0:User"
    ]

    employee_number: str | None = None
    """Numeric or alphanumeric identifier assigned to a person, typically based
    on order of hire or association with an organization."""

    organization: str | None = None
    """Identifies the name of an organization."""

    department: str | None = None
    """Identifies the name of a department."""


def get_service_provider_config():
    return ServiceProviderConfig(
        meta=Meta(
            resource_type="ServiceProviderConfig",
            location=url_for("scim.query_service_provider_config", _external=True),
        ),
        documentation_uri=DOCUMENTATION_URL,
        patch=Patch(supported=True),
        bulk=Bulk(supported=False, max_operations=0, max_payload_size=0),
        change_password=ChangePassword(supported=True),
        filter=Filter(supported=False, max_results=0),
        sort=Sort(supported=False),
        etag=ETag(supported=False),
        authentication_schemes=[
            AuthenticationScheme(
                name="OAuth Bearer Token",
                description="Authentication scheme using the OAuth Bearer Token Standard",
                spec_uri="http://www.rfc-editor.org/info/rfc6750",
                documentation_uri=DOCUMENTATION_URL,
                type="oauthbearertoken",
                primary=True,
            ),
        ],
    )


def get_resource_types():
    """Return the resource types implemented by Canaille."""
    return {
        "User": ResourceType(
            id="User",
            name="User",
            endpoint=url_for("scim.query_users", _external=True),
            description="User accounts",
            schema_="urn:ietf:params:scim:schemas:core:2.0:User",
            schema_extensions=[
                SchemaExtension(
                    schema_="urn:ietf:params:scim:schemas:extension:enterprise:2.0:User",
                    required=True,
                )
            ],
            meta=Meta(
                resource_type="ResourceType",
                location=url_for(
                    "scim.query_resource_type",
                    resource_type_name="User",
                    _external=True,
                ),
            ),
        ),
        "Group": ResourceType(
            id="Group",
            name="Group",
            endpoint=url_for("scim.query_groups", _external=True),
            description="Group management",
            schema_="urn:ietf:params:scim:schemas:core:2.0:Group",
            meta=Meta(
                resource_type="ResourceType",
                location=url_for(
                    "scim.query_resource_type",
                    resource_type_name="Group",
                    _external=True,
                ),
            ),
        ),
    }


def get_schemas():
    schemas = {
        model.model_fields["schemas"].default[0]: model.to_schema()
        for model in (
            ServiceProviderConfig,
            ResourceType,
            Schema,
            User,
            Group,
            EnterpriseUser,
        )
    }
    for schema_id, schema in schemas.items():
        schema.meta = Meta(
            resource_type="Schema",
            location=url_for("scim.query_schema", schema_id=schema_id, _external=True),
        )
    return schemas



import wtforms.form
import wtforms.validators
from flask import current_app

from canaille.app import models
from canaille.app.i18n import gettext
from canaille.app.i18n import lazy_gettext as _
from canaille.backends import Backend
from canaille.core.auth import get_user_from_login


def unique_user_name(form, field):
    if Backend.instance.get(models.User, user_name=field.data) and (
        not getattr(form, "user", None) or form.user.user_name != field.data
    ):
        raise wtforms.ValidationError(
            _("The user name '{user_name}' already exists").format(user_name=field.data)
        )


def unique_email(form, field):
    if Backend.instance.get(models.User, emails=field.data) and (
        not getattr(form, "user", None) or field.data not in (form.user.emails or [])
    ):
        raise wtforms.ValidationError(
            _("The email '{email}' is already used").format(email=field.data)
        )


def unique_group(form, field):
    if Backend.instance.get(models.Group, display_name=field.data):
        raise wtforms.ValidationError(
            _("The group '{group}' already exists").format(group=field.data)
        )


def existing_login(form, field):
    if not current_app.config["CANAILLE"][
        "HIDE_INVALID_LOGINS"
    ] and not get_user_from_login(field.data):
        raise wtforms.ValidationError(
            _("The login '{login}' does not exist").format(login=field.data)
        )


def existing_group_member(form, field):
    if field.data is None:
        raise wtforms.ValidationError(
            gettext("The user you are trying to remove does not exist.")
        )

    if field.data not in form.group.members:
        raise wtforms.ValidationError(
            gettext(
                "The user '{user}' has already been removed from the group '{group}'"
            ).format(user=field.data.formatted_name, group=form.group.display_name)
        )


def non_empty_groups(form, field):
    """LDAP groups cannot be empty because groupOfNames.member is a MUST attribute.

    https://www.rfc-editor.org/rfc/rfc2256.html#section-7.10
    """
    if not form.user:
        return

    for group in form.user.groups:
        if len(group.members) == 1 and group not in field.data:
            raise wtforms.ValidationError(
                _(
                    "The group '{group}' cannot be removed, because it must have at least one user left."
                ).format(group=group.display_name)
            )



def test_json_oauth_errors(testclient):
    """Checks that HTTP errors on the oauth endpoints are in the JSON format."""
    res = testclient.get("/oauth/invalid", status=404)
    assert res.json == {
        "error": "not_found",
        "error_description": "The requested URL was not found on the server. If you entered the URL manually please check your spelling and try again.",
    }


def test_missing_client_id(testclient, logged_user, client, trusted_client, backend):
    """Missing client_id should raise a 400 error."""
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            scope="openid profile email groups address phone",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
        status=400,
    )
    assert "Missing 'client_id' parameter." in str(res.html)


def test_invalid_client(
    testclient,
    logged_user,
):
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id="invalid",
            scope="openid profile email groups address phone",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
        status=400,
    )
    res.mustcontain("The client does not exist on this server.")


def test_no_redirect_uri(testclient, logged_user, client, backend):
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            nonce="somenonce",
        ),
        status=400,
    )
    assert "Missing 'redirect_uri' in request." in str(res.html)


def test_invalid_redirect_uri(testclient, logged_user, client, backend):
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            nonce="somenonce",
            redirect_uri="https://invalid.test",
        ),
        status=400,
    )
    assert "Redirect URI https://invalid.test is not supported by client." in str(
        res.html
    )



"""Remove client.trusted column.

Revision ID: 1758051913
Revises: 1755869767
Create Date: 2025-01-17 08:58:33.000000

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1758051913"
down_revision = "1755869767"
branch_labels = None
depends_on = None


def upgrade():
    """Remove the trusted column from client table."""
    with op.batch_alter_table("client") as batch_op:
        batch_op.drop_column("trusted")


def downgrade():
    """Re-add the trusted column to client table."""
    with op.batch_alter_table("client") as batch_op:
        batch_op.add_column(sa.Column("trusted", sa.Boolean(), nullable=True))



import json
from http import HTTPStatus

from authlib.integrations.flask_oauth2 import ResourceProtector
from authlib.integrations.flask_oauth2.errors import (
    _HTTPException as AuthlibHTTPException,
)
from authlib.oauth2.rfc6750 import BearerTokenValidator
from flask import Blueprint
from flask import Response
from flask import abort
from flask import request
from pydantic import ValidationError
from scim2_models import Context
from scim2_models import Error
from scim2_models import ListResponse
from scim2_models import PatchOp
from scim2_models import ResourceType
from scim2_models import Schema
from scim2_models import SearchRequest
from werkzeug.exceptions import HTTPException

from canaille.app import models
from canaille.app.flask import csrf
from canaille.backends import Backend

from .casting import group_from_canaille_to_scim_server
from .casting import group_from_scim_to_canaille
from .casting import user_from_canaille_to_scim_server
from .casting import user_from_scim_to_canaille
from .models import EnterpriseUser
from .models import Group
from .models import User
from .models import get_resource_types
from .models import get_schemas
from .models import get_service_provider_config

bp = Blueprint("scim", __name__, url_prefix="/scim/v2")


class SCIMBearerTokenValidator(BearerTokenValidator):
    def authenticate_token(self, token_string: str):
        token = Backend.instance.get(models.Token, access_token=token_string)
        # At the moment, only client tokens are allowed, and not user tokens
        return token if token and not token.subject else None


require_oauth = ResourceProtector()
require_oauth.register_token_validator(SCIMBearerTokenValidator())


@bp.after_request
def add_scim_content_type(response):
    response.headers["Content-Type"] = "application/scim+json"
    return response


@bp.errorhandler(HTTPException)
def http_error_handler(error):
    obj = Error(detail=str(error), status=error.code)
    return obj.model_dump(), obj.status


@bp.errorhandler(AuthlibHTTPException)
def oauth2_error(error):
    body = json.loads(error.body)
    obj = Error(
        detail=f"{body['error']}: {body['error_description']}"
        if "error_description" in body
        else body["error"],
        status=error.code,
    )
    return obj.model_dump(), error.code


@bp.errorhandler(ValidationError)
def scim_error_handler(error):
    error_details = error.errors()[0]
    obj = Error(
        status=400,
        detail=f"{error_details['msg']}: {' ,'.join(str(loc) for loc in error_details['loc'])}",
    )
    # TODO: maybe the Pydantic <=> SCIM error code mapping could go in scim2_models
    obj.scim_type = (
        "invalidValue" if error_details["type"] == "required_error" else None
    )

    return obj.model_dump(), obj.status


def parse_search_request(request) -> SearchRequest:
    """Create a SearchRequest object from the request arguments."""
    max_nb_items_per_page = 1000
    count = (
        min(request.args["count"], max_nb_items_per_page)
        if request.args.get("count")
        else None
    )
    req = SearchRequest(
        attributes=request.args.get("attributes"),
        excluded_attributes=request.args.get("excludedAttributes"),
        start_index=request.args.get("startIndex"),
        count=count,
    )
    return req


@bp.route("/Users", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_users():
    req = parse_search_request(request)
    users = list(
        Backend.instance.query(models.User)[req.start_index_0 : req.stop_index_0]
    )
    total = len(users)
    scim_users = [user_from_canaille_to_scim_server(user) for user in users]
    list_response = ListResponse[User[EnterpriseUser]](
        start_index=req.start_index,
        items_per_page=req.count,
        total_results=total,
        resources=scim_users,
    )
    payload = list_response.model_dump(
        scim_ctx=Context.RESOURCE_QUERY_RESPONSE,
    )
    return payload


@bp.route("/Users/<user:user>", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_user(user):
    scim_user = user_from_canaille_to_scim_server(user)
    return scim_user.model_dump(
        scim_ctx=Context.RESOURCE_QUERY_RESPONSE,
    )


@bp.route("/Groups", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_groups():
    req = parse_search_request(request)
    groups = list(
        Backend.instance.query(models.group)[req.start_index_0 : req.stop_index_0]
    )
    total = len(groups)
    scim_groups = [group_from_canaille_to_scim_server(group) for group in groups]
    list_response = ListResponse[Group](
        start_index=req.start_index,
        items_per_page=req.count,
        total_results=total,
        resources=scim_groups,
    )
    payload = list_response.model_dump(
        scim_ctx=Context.RESOURCE_QUERY_RESPONSE,
    )
    return payload


@bp.route("/Groups/<group:group>", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_group(group):
    scim_group = group_from_canaille_to_scim_server(group)
    return scim_group.model_dump(
        scim_ctx=Context.RESOURCE_QUERY_RESPONSE,
    )


@bp.route("/Schemas", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_schemas():
    req = parse_search_request(request)
    schemas = list(get_schemas().values())[req.start_index_0 : req.stop_index_0]
    response = ListResponse[Schema](
        total_results=len(schemas),
        items_per_page=req.count or len(schemas),
        start_index=req.start_index,
        resources=schemas,
    )
    return response.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/Schemas/<string:schema_id>", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_schema(schema_id):
    schema = get_schemas().get(schema_id)
    if not schema:
        abort(404)

    return schema.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/ResourceTypes", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_resource_types():
    req = parse_search_request(request)
    resource_types = list(get_resource_types().values())[
        req.start_index_0 : req.stop_index_0
    ]
    response = ListResponse[ResourceType](
        total_results=len(resource_types),
        items_per_page=req.count or len(resource_types),
        start_index=req.start_index,
        resources=resource_types,
    )
    return response.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/ResourceTypes/<string:resource_type_name>", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_resource_type(resource_type_name):
    resource_type = get_resource_types().get(resource_type_name)
    if not resource_type:
        abort(404)

    return resource_type.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/ServiceProviderConfig", methods=["GET"])
@csrf.exempt
@require_oauth()
def query_service_provider_config():
    spc = get_service_provider_config()
    return spc.model_dump(scim_ctx=Context.RESOURCE_QUERY_RESPONSE)


@bp.route("/Users", methods=["POST"])
@csrf.exempt
@require_oauth()
def create_user():
    request_user = User[EnterpriseUser].model_validate(
        request.json, scim_ctx=Context.RESOURCE_CREATION_REQUEST
    )
    user = user_from_scim_to_canaille(request_user, models.User())
    Backend.instance.save(user)
    response_user = user_from_canaille_to_scim_server(user)
    payload = response_user.model_dump_json(scim_ctx=Context.RESOURCE_CREATION_RESPONSE)
    return Response(payload, status=HTTPStatus.CREATED)


@bp.route("/Groups", methods=["POST"])
@csrf.exempt
@require_oauth()
def create_group():
    request_group = Group.model_validate(
        request.json, scim_ctx=Context.RESOURCE_CREATION_REQUEST
    )
    group = group_from_scim_to_canaille(request_group, models.Group())
    Backend.instance.save(group)
    response_group = group_from_canaille_to_scim_server(group)
    payload = response_group.model_dump_json(
        scim_ctx=Context.RESOURCE_CREATION_RESPONSE
    )
    return Response(payload, status=HTTPStatus.CREATED)


@bp.route("/Users/<user:user>", methods=["PUT"])
@csrf.exempt
@require_oauth()
def replace_user(user):
    original_scim_user = user_from_canaille_to_scim_server(user)
    request_scim_user = User[EnterpriseUser].model_validate(
        request.json,
        scim_ctx=Context.RESOURCE_REPLACEMENT_REQUEST,
        original=original_scim_user,
    )
    updated_user = user_from_scim_to_canaille(request_scim_user, user)
    Backend.instance.save(updated_user)
    response_scim_user = user_from_canaille_to_scim_server(updated_user)
    payload = response_scim_user.model_dump(
        scim_ctx=Context.RESOURCE_REPLACEMENT_RESPONSE
    )
    return payload


@bp.route("/Groups/<group:group>", methods=["PUT"])
@csrf.exempt
@require_oauth()
def replace_group(group):
    original_scim_group = group_from_canaille_to_scim_server(group)
    request_scim_group = Group.model_validate(
        request.json,
        scim_ctx=Context.RESOURCE_REPLACEMENT_REQUEST,
        original=original_scim_group,
    )
    updated_group = group_from_scim_to_canaille(request_scim_group, group)
    Backend.instance.save(updated_group)
    response_group = group_from_canaille_to_scim_server(updated_group)
    payload = response_group.model_dump(scim_ctx=Context.RESOURCE_REPLACEMENT_RESPONSE)
    return payload


@bp.route("/Users/<user:user>", methods=["PATCH"])
@csrf.exempt
@require_oauth()
def patch_user(user):
    scim_user = user_from_canaille_to_scim_server(user)
    patch_op = PatchOp[User[EnterpriseUser]].model_validate(
        request.json, scim_ctx=Context.RESOURCE_PATCH_REQUEST
    )
    modified = patch_op.patch(scim_user)

    if modified:
        updated_user = user_from_scim_to_canaille(scim_user, user)
        Backend.instance.save(updated_user)
        scim_user = user_from_canaille_to_scim_server(updated_user)

    return scim_user.model_dump(scim_ctx=Context.RESOURCE_PATCH_RESPONSE)


@bp.route("/Groups/<group:group>", methods=["PATCH"])
@csrf.exempt
@require_oauth()
def patch_group(group):
    scim_group = group_from_canaille_to_scim_server(group)
    patch_op = PatchOp[Group].model_validate(
        request.json, scim_ctx=Context.RESOURCE_PATCH_REQUEST
    )
    modified = patch_op.patch(scim_group)

    if modified:
        updated_group = group_from_scim_to_canaille(scim_group, group)
        Backend.instance.save(updated_group)
        scim_group = group_from_canaille_to_scim_server(updated_group)

    return scim_group.model_dump(scim_ctx=Context.RESOURCE_PATCH_RESPONSE)


@bp.route("/Users/<user:user>", methods=["DELETE"])
@csrf.exempt
@require_oauth()
def delete_user(user):
    Backend.instance.delete(user)
    return "", HTTPStatus.NO_CONTENT


@bp.route("/Groups/<group:group>", methods=["DELETE"])
@csrf.exempt
@require_oauth()
def delete_group(group):
    Backend.instance.delete(group)
    return "", HTTPStatus.NO_CONTENT



import datetime
from typing import ClassVar

from authlib.oauth2.rfc6749 import ClientMixin
from authlib.oauth2.rfc6749 import TokenMixin
from authlib.oauth2.rfc6749 import util
from authlib.oidc.core import AuthorizationCodeMixin
from blinker import signal

from canaille.app import models
from canaille.backends import Backend

from .basemodels import AuthorizationCode as BaseAuthorizationCode
from .basemodels import Client as BaseClient
from .basemodels import Consent as BaseConsent
from .basemodels import Token as BaseToken


class Client(BaseClient, ClientMixin):
    client_info_attributes: ClassVar[list[str]] = [
        "client_id",
        "client_secret",
        "client_id_issued_at",
        "client_secret_expires_at",
    ]

    client_metadata_attributes: ClassVar[list[str]] = [
        "client_name",
        "contacts",
        "client_uri",
        "redirect_uris",
        "logo_uri",
        "grant_types",
        "response_types",
        "scope",
        "tos_uri",
        "policy_uri",
        "jwks_uri",
        "jwks",
        "token_endpoint_auth_method",
        "token_endpoint_auth_signing_alg",
        "software_id",
        "software_version",
        "sector_identifier_uri",
        "subject_type",
        "application_type",
        "id_token_signed_response_alg",
        "id_token_encrypted_response_alg",
        "id_token_encrypted_response_enc",
        "userinfo_signed_response_alg",
        "userinfo_encrypted_response_alg",
        "userinfo_encrypted_response_enc",
        "default_max_age",
        "require_auth_time",
        "default_acr_values",
        "initiate_login_uri",
        "request_object_signing_alg",
        "request_object_encryption_alg",
        "request_object_encryption_enc",
        "request_uris",
        "require_signed_request_object",
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        signal("before_client_delete").connect(self.on_delete, sender=self)

    def get_client_id(self):
        return self.client_id

    def get_allowed_scope(self, scope: str) -> str:
        allowed_scope = [
            scope_piece
            for scope_piece in (self.scope or scope.split())
            if scope_piece in scope
        ]
        return util.list_to_scope(allowed_scope)

    def check_redirect_uri(self, redirect_uri: str) -> bool:
        return redirect_uri in self.redirect_uris

    def check_client_secret(self, client_secret: str) -> bool:
        return client_secret == self.client_secret

    def check_endpoint_auth_method(self, method, endpoint):
        if endpoint == "token":
            # hotfix for https://github.com/lepture/authlib/issues/759
            implicit_flow_shortcut = method == "none" and self.grant_types == [
                "implicit"
            ]
            return (
                implicit_flow_shortcut
                or method == self.token_endpoint_auth_method
                or (
                    method == "client_assertion_jwt"
                    and self.token_endpoint_auth_method
                    in ("client_secret_jwt", "private_key_jwt")
                )
            )
        return True

    def check_response_type(self, response_type):
        return all(r in self.response_types for r in response_type.split(" "))

    def check_grant_type(self, grant_type):
        return grant_type in self.grant_types

    @property
    def client_info(self):
        result = {
            attribute_name: getattr(self, attribute_name)
            for attribute_name in self.client_info_attributes
        }
        result["client_id_issued_at"] = int(
            datetime.datetime.timestamp(result["client_id_issued_at"])
        )
        result["client_secret_expires_at"] = (
            int(datetime.datetime.timestamp(result["client_secret_expires_at"]))
            if result["client_secret_expires_at"]
            else 0
        )
        return result

    @property
    def client_metadata(self):
        metadata = {
            attribute_name: getattr(self, attribute_name)
            for attribute_name in self.client_metadata_attributes
        }
        metadata["scope"] = " ".join(metadata["scope"])
        return metadata

    @classmethod
    def on_delete(cls, self, data):
        for consent in Backend.instance.query(models.Consent, client=self):
            Backend.instance.delete(consent)

        for code in Backend.instance.query(models.AuthorizationCode, client=self):
            Backend.instance.delete(code)

        for token in Backend.instance.query(models.Token, client=self):
            Backend.instance.delete(token)

    def get_default_redirect_uri(self):
        return None


class AuthorizationCode(BaseAuthorizationCode, AuthorizationCodeMixin):
    def get_redirect_uri(self):
        return self.redirect_uri

    def get_scope(self) -> str:
        return util.list_to_scope(self.scope)

    def get_nonce(self):
        return self.nonce

    def is_expired(self):
        return self.issue_date + datetime.timedelta(
            seconds=int(self.lifetime)
        ) < datetime.datetime.now(datetime.timezone.utc)

    def get_auth_time(self) -> float | None:
        return self.auth_time.timestamp() if self.auth_time else None

    def get_acr(self):
        return self.acr

    def get_amr(self):
        return self.amr


class Token(BaseToken, TokenMixin):
    @property
    def expire_date(self):
        return self.issue_date + datetime.timedelta(seconds=self.lifetime)

    @property
    def revoked(self):
        return bool(self.revokation_date)

    def get_scope(self):
        return " ".join(self.scope)

    def get_issued_at(self) -> float:
        return self.issue_date.timestamp()

    def get_expires_at(self) -> float:
        return self.get_issued_at() + self.lifetime

    def is_refresh_token_active(self):
        if self.revokation_date:
            return False

        return self.expire_date >= datetime.datetime.now(datetime.timezone.utc)

    def is_expired(self):
        return self.issue_date + datetime.timedelta(
            seconds=int(self.lifetime)
        ) < datetime.datetime.now(datetime.timezone.utc)

    def is_revoked(self):
        return bool(self.revokation_date)

    def check_client(self, client):
        return client.client_id == self.client.client_id

    def get_user(self):
        return self.subject

    def get_client(self):
        return self.client


class Consent(BaseConsent):
    @property
    def revoked(self):
        return bool(self.revokation_date)

    def revoke(self):
        self.revokation_date = datetime.datetime.now(datetime.timezone.utc)
        Backend.instance.save(self)

        tokens = Backend.instance.query(
            models.Token,
            client=self.client,
            subject=self.subject,
        )
        tokens = [token for token in tokens if not token.revoked]
        for t in tokens:
            t.revokation_date = self.revokation_date
            Backend.instance.save(t)

    def restore(self):
        self.revokation_date = None
        Backend.instance.save(self)



import datetime
import uuid

import pytest
from httpx import Client as httpx_client
from scim2_client.engines.httpx import SyncSCIMClient
from scim2_client.engines.werkzeug import TestSCIMClient
from scim2_models import Resource
from werkzeug.security import gen_salt
from werkzeug.test import Client

from canaille.app import models
from canaille.scim.endpoints import bp
from canaille.scim.endpoints import get_resource_types
from canaille.scim.endpoints import get_schemas
from canaille.scim.endpoints import get_service_provider_config


@pytest.fixture
def configuration(configuration):
    configuration["CANAILLE_SCIM"] = {
        "ENABLE_SERVER": True,
        "ENABLE_CLIENT": True,
    }
    return configuration


@pytest.fixture
def oidc_client(testclient, backend):
    c = models.Client(
        client_id=gen_salt(24),
        client_name="Some client",
        contacts=["contact@mydomain.test"],
        client_uri="https://mydomain.test",
        redirect_uris=[
            "https://mydomain.test/redirect1",
        ],
        client_id_issued_at=datetime.datetime.now(datetime.timezone.utc),
        client_secret=gen_salt(48),
        grant_types=[
            "client_credentials",
        ],
        response_types=["code", "token", "id_token"],
        scope=["openid", "email", "profile", "groups", "address", "phone"],
        token_endpoint_auth_method="client_secret_basic",
    )
    backend.save(c)

    yield c
    backend.delete(c)


@pytest.fixture
def oidc_token(testclient, oidc_client, backend):
    t = models.Token(
        token_id=gen_salt(48),
        access_token=gen_salt(48),
        audience=[oidc_client],
        client=oidc_client,
        refresh_token=gen_salt(48),
        scope=["openid", "profile"],
        issue_date=datetime.datetime.now(datetime.timezone.utc),
        lifetime=3600,
    )
    backend.save(t)
    yield t
    backend.delete(t)


@pytest.fixture
def scim_client(app, oidc_client, oidc_token):
    resource_models = [
        Resource.from_schema(schema) for schema in get_schemas().values()
    ]

    return TestSCIMClient(
        Client(app),
        scim_prefix=bp.url_prefix,
        environ={"headers": {"Authorization": f"Bearer {oidc_token.access_token}"}},
        check_response_status_codes=False,
        service_provider_config=get_service_provider_config(),
        resource_types=get_resource_types().values(),
        resource_models=resource_models,
    )


@pytest.fixture
def scim_trusted_client(testclient, scim2_server, backend):
    client_uri = f"http://localhost:{scim2_server.port}"
    c = models.Client(
        client_id=gen_salt(24),
        client_name="Some client",
        contacts=["contact@mydomain.test"],
        client_uri=client_uri,
        redirect_uris=[
            client_uri + "/redirect1",
        ],
        client_id_issued_at=datetime.datetime.now(datetime.timezone.utc),
        client_secret=gen_salt(48),
        grant_types=[
            "client_credentials",
        ],
        response_types=["code", "token", "id_token"],
        scope=["openid", "email", "profile", "groups", "address", "phone"],
        token_endpoint_auth_method="client_secret_basic",
    )
    backend.save(c)
    yield c
    backend.delete(c)


@pytest.fixture
def scim_token(testclient, scim_trusted_client, backend):
    t = models.Token(
        token_id=gen_salt(48),
        access_token=gen_salt(48),
        audience=[scim_trusted_client],
        client=scim_trusted_client,
        refresh_token=gen_salt(48),
        scope=["openid", "profile"],
        issue_date=datetime.datetime.now(datetime.timezone.utc),
        lifetime=3600,
    )
    backend.save(t)
    yield t
    backend.delete(t)


@pytest.fixture
def scim_client_for_trusted_client(scim2_server, scim_token):
    client_httpx = httpx_client(
        base_url=f"http://localhost:{scim2_server.port}",
        headers={"Authorization": f"Bearer {scim_token.access_token}"},
    )
    scim_client = SyncSCIMClient(client_httpx)
    scim_client.discover()
    return scim_client


@pytest.fixture
def client_without_scim(testclient, backend):
    c = models.Client(
        client_id=gen_salt(24),
        client_name="Client",
        contacts=["contact@trusted.test"],
        client_uri="https://untrusted.test",
        redirect_uris=[
            "https://client.trusted.test/redirect1",
            "https://client.trusted.test/redirect2",
        ],
        logo_uri="https://client.trusted.test/logo.webp",
        client_id_issued_at=datetime.datetime.now(datetime.timezone.utc),
        client_secret=gen_salt(48),
        grant_types=[
            "password",
            "authorization_code",
            "implicit",
            "hybrid",
            "refresh_token",
            "client_credentials",
        ],
        response_types=["code", "token", "id_token"],
        scope=["openid", "profile", "groups"],
        tos_uri="https://client.trusted.test/tos",
        policy_uri="https://client.trusted.test/policy",
        jwks_uri="https://client.trusted.test/jwk",
        token_endpoint_auth_method="client_secret_basic",
        post_logout_redirect_uris=["https://client.trusted.test/disconnected"],
    )
    backend.save(c)
    c.audience = [c]
    backend.save(c)

    yield c
    backend.delete(c)


@pytest.fixture
def consent(testclient, client_without_scim, user, backend):
    t = models.Consent(
        consent_id=str(uuid.uuid4()),
        client=client_without_scim,
        subject=user,
        scope=["openid", "profile"],
        issue_date=datetime.datetime.now(datetime.timezone.utc),
    )
    backend.save(t)
    yield t
    backend.delete(t)



import datetime
import logging

import otpauth
import pytest
import time_machine

from canaille.app import models
from canaille.app.otp import HOTP_LOOK_AHEAD_WINDOW
from canaille.app.otp import HOTP_START_COUNTER
from canaille.core.auth import AuthenticationSession


@pytest.fixture
def configuration(configuration):
    configuration["CANAILLE"]["AUTHENTICATION_FACTORS"] = ["otp"]
    return configuration


def generate_otp(method, secret, hotp_counter=None):
    hotp_counter = HOTP_START_COUNTER if hotp_counter is None else hotp_counter
    if method == "TOTP":
        totp = otpauth.TOTP(secret.encode("utf-8"))
        return totp.string_code(totp.generate())

    else:
        hotp = otpauth.HOTP(secret.encode("utf-8"))
        return hotp.string_code(hotp.generate(hotp_counter))


def test_no_auth_session_not_logged_in(testclient, user):
    """Non-logged users should not be able to access the otp auth form if an authentication session has not been properly started."""
    res = testclient.get("/auth/otp", status=302)
    assert res.location == "/login"
    assert res.flashes == [
        ("warning", "Cannot remember the login you attempted to sign in with.")
    ]


def test_no_auth_session_logged_in(testclient, logged_user):
    """Logged users should not be able to access the otp auth form if an authentication session has not been properly started."""
    res = testclient.get("/auth/otp", status=302)
    assert res.location == "/"


def test_not_otp_step(testclient, user):
    """Users reaching the otp form while this is not the right auth step in their flow should be redirected there."""
    with testclient.session_transaction() as session:
        session["auth"] = AuthenticationSession(
            user_name="user",
            remaining=["email", "otp"],
        ).serialize()

    res = testclient.get("/auth/otp", status=302)
    assert res.location == "/auth/email"


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_signin_with_otp(testclient, user, caplog, otp_method):
    """Nominal case for OTP authentication."""
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["otp"] = generate_otp(otp_method, user.secret_token, user.hotp_counter)
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome Johnny",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Successful authentication for user",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_signin_wrong_otp(testclient, user, caplog, otp_method):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["otp"] = 111111
    res = res.form.submit()

    assert (
        "error",
        "The passcode you entered is invalid. Please try again",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed OTP authentication for user",
    ) in caplog.record_tuples


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_signin_wrong_user(testclient, user, caplog, otp_method):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "invalid"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["otp"] = 111111
    res = res.form.submit()

    assert (
        "error",
        "The passcode you entered is invalid. Please try again",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed OTP authentication for invalid",
    ) in caplog.record_tuples


def test_signin_expired_totp(testclient, user, caplog):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = "TOTP"

    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False) as traveller:
        with testclient.session_transaction() as session:
            assert not session.get("sessions")

        res = testclient.get("/login", status=200)

        res.form["login"] = "user"
        res = res.form.submit(status=302)
        res = res.follow(status=200)

        res.form["otp"] = generate_otp("TOTP", user.secret_token, user.hotp_counter)
        traveller.shift(datetime.timedelta(seconds=30))
        res = res.form.submit()

        assert (
            "error",
            "The passcode you entered is invalid. Please try again",
        ) in res.flashes
        assert (
            "canaille",
            logging.SECURITY,
            "Failed OTP authentication for user",
        ) in caplog.record_tuples


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_signin_invalid_otp_chars(testclient, user, caplog, otp_method):
    """Test putting unicode characters in the OTP field."""
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["otp"] = "●●●●●●"
    res = res.form.submit()

    assert (
        "error",
        "The passcode you entered is invalid. Please try again",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed OTP authentication for user",
    ) in caplog.record_tuples


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_setup_otp(testclient, backend, caplog, otp_method):
    """Users who have never logged in with OTP should be shown the OTP setup screen."""
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    user = models.User(
        formatted_name="Otp User",
        family_name="Otp",
        user_name="otp",
        emails=["john@doe.test"],
        password="correct horse battery staple",
    )
    backend.save(user)

    if otp_method == "HOTP":
        assert user.hotp_counter is None

    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "otp"
    res = res.form.submit(status=302)
    res = res.follow(status=302)

    assert res.location == "/auth/otp-setup"
    res = res.follow(status=200)

    with testclient.session_transaction() as session:
        secret_token = (
            session["auth"]["data"]["otp_user_secret"]
            if "data" in session["auth"]
            else None
        )

    res.form["otp"] = generate_otp(otp_method, secret_token, user.hotp_counter)
    res = res.form.submit(status=302)

    backend.reload(user)
    if otp_method == "HOTP":
        assert user.hotp_counter == 1

    assert (
        "success",
        "Connection successful. Welcome Otp User",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Successful authentication for otp",
    ) in caplog.record_tuples
    res = res.follow(status=302)
    res = res.follow(status=200)

    backend.delete(user)


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_setup_otp_bad_otp(testclient, backend, caplog, otp_method):
    """Users who have never logged in with OTP should be shown the OTP setup screen."""
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    user = models.User(
        formatted_name="Otp User",
        family_name="Otp",
        user_name="otp",
        emails=["john@doe.test"],
        password="correct horse battery staple",
    )
    backend.save(user)

    if otp_method == "HOTP":
        assert user.hotp_counter is None

    with testclient.session_transaction() as session:
        assert not session.get("sessions")

    res = testclient.get("/login", status=200)

    res.form["login"] = "otp"
    res = res.form.submit(status=302)
    res = res.follow(status=302)

    assert res.location == "/auth/otp-setup"
    res = res.follow(status=200)

    res.form["otp"] = 111111
    res = res.form.submit(status=200)

    assert (
        "error",
        "The passcode you entered is invalid. Please try again.",
    ) in res.flashes

    backend.reload(user)
    assert not user.secret_token

    backend.delete(user)


def test_signin_multiple_attempts_doesnt_desynchronize_hotp(testclient, user, caplog):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = "HOTP"

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    for _ in range(3):
        res.form["otp"] = "111111"
        res = res.form.submit(status=200)

    res.form["otp"] = generate_otp("HOTP", user.secret_token, user.hotp_counter)
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome Johnny",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Successful authentication for user",
    ) in caplog.record_tuples


@pytest.mark.parametrize("otp_method", ["TOTP", "HOTP"])
def test_setup_otp_page_already_logged_in(testclient, logged_user, otp_method):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = otp_method

    res = testclient.get("/auth/otp-setup", status=302)
    assert res.location == "/profile/user/settings"


def test_signin_inside_hotp_look_ahead_window(testclient, backend, user, caplog):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = "HOTP"

    assert user.hotp_counter == 1

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["otp"] = generate_otp(
        "HOTP", user.secret_token, user.hotp_counter + HOTP_LOOK_AHEAD_WINDOW
    )

    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome Johnny",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Successful authentication for user",
    ) in caplog.record_tuples

    user = backend.get(models.User, id=user.id)
    assert user.hotp_counter == HOTP_LOOK_AHEAD_WINDOW + 2


def test_signin_outside_hotp_look_ahead_window(testclient, backend, user, caplog):
    testclient.app.config["CANAILLE"]["OTP_METHOD"] = "HOTP"

    assert user.hotp_counter == 1

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["otp"] = generate_otp(
        "HOTP", user.secret_token, user.hotp_counter + HOTP_LOOK_AHEAD_WINDOW + 1
    )
    res = res.form.submit(status=200)

    assert (
        "error",
        "The passcode you entered is invalid. Please try again",
    ) in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        "Failed OTP authentication for user",
    ) in caplog.record_tuples

    user = backend.get(models.User, id=user.id)
    assert user.hotp_counter == 1



"""Methods related to HOTP and TOTP."""

import secrets

from flask import current_app

SECRET_TOKEN_LENGTH = 8
HOTP_LOOK_AHEAD_WINDOW = 10
HOTP_START_COUNTER = 0


def make_otp_secret():
    return secrets.token_hex(SECRET_TOKEN_LENGTH)


def get_otp_authentication_setup_uri(user, secret):
    import otpauth

    method = current_app.features.otp_method
    if method == "TOTP":
        otp = otpauth.TOTP(secret.encode("utf-8"))
        return otp.to_uri(
            label=user.user_name, issuer=current_app.config["CANAILLE"]["NAME"]
        ), otp.b32_secret

    elif method == "HOTP":
        otp = otpauth.HOTP(secret.encode("utf-8"))
        return otp.to_uri(
            label=user.user_name,
            issuer=current_app.config["CANAILLE"]["NAME"],
            counter=HOTP_START_COUNTER,
        ), otp.b32_secret

    else:  # pragma: no cover
        raise RuntimeError("Invalid one-time passcode method")


def is_otp_valid(
    otp, method, secret_token, hotp_counter=HOTP_START_COUNTER
) -> tuple[bool, int]:
    """Return whether the OTP code is valid, and the new hotp_counter for the user."""
    import otpauth

    if method == "TOTP":
        result = otpauth.TOTP(secret_token.encode("utf-8")).verify(otp)
        return result, hotp_counter

    else:
        # if user token's counter is ahead of canaille's, try to catch up to it
        for counter in range(hotp_counter, hotp_counter + HOTP_LOOK_AHEAD_WINDOW + 1):
            if otpauth.HOTP(secret_token.encode("utf-8")).verify(otp, counter):
                return True, counter + 1
        return False, hotp_counter + 1



import sys

from flask import Flask


def create_app(
    config: dict | None = None,
    backend=None,
    init_backend=None,
    env_file: str | None = None,
    env_prefix: str = "",
):
    """Application entry point.

    :param config: A configuration dict. This will take priority over any other configuration method.
    :param backend: An optional backend to force. If unset backend will be initialized according to the configuration.
    :param env_file: The path to an environment var file in which configuration can be loaded.
    :param env_prefix: The prefix to configuration environment vars.
    """
    from .app.configuration import setup_config
    from .app.features import setup_features
    from .app.flask import setup_flask
    from .app.flask import setup_flask_blueprints
    from .app.flask import setup_flask_converters
    from .app.i18n import setup_i18n
    from .app.logging import setup_logging
    from .app.security import setup_security
    from .app.sentry import setup_sentry
    from .app.templating import setup_jinja
    from .app.templating import setup_themer
    from .backends import setup_backend

    app = Flask(__name__)
    with app.app_context():
        setup_features(app)
        if not setup_config(
            app=app,
            config=config,
            env_file=env_file,
            env_prefix=env_prefix,
        ):  # pragma: no cover
            sys.exit(1)

    sentry_sdk = setup_sentry(app)
    try:
        setup_logging(app)

        backend = setup_backend(app, backend, init_backend)
        setup_security(app)
        setup_flask_converters(app)
        setup_flask_blueprints(app)
        setup_jinja(app)
        setup_i18n(app)
        setup_themer(app)
        setup_flask(app)

        if app.features.has_oidc:  # pragma: no branch
            from .oidc.oauth import setup_oauth

            setup_oauth(app)

        if app.features.has_scim_client:
            from .scim.client import setup_scim_client

            setup_scim_client()

    except Exception as exc:  # pragma: no cover
        if sentry_sdk:
            sentry_sdk.capture_exception(exc)
        raise

    return app



import pytest
from flask_webtest import TestApp

from canaille import create_app


@pytest.fixture
def configuration(slapd_server, ldap_configuration):
    yield ldap_configuration


def test_ldap_connection_remote_ldap_unreachable(configuration):
    app = create_app(configuration)
    testclient = TestApp(app)

    app.config["CANAILLE_LDAP"]["URI"] = "ldap://invalid-ldap.com"

    app.config["DEBUG"] = False
    res = testclient.get("/", status=500, expect_errors=True)
    res.mustcontain(no="Could not connect to the LDAP server")


def test_ldap_connection_remote_ldap_wrong_credentials(configuration):
    app = create_app(configuration)
    testclient = TestApp(app)

    app.config["CANAILLE_LDAP"]["BIND_PW"] = "invalid-password"

    app.config["DEBUG"] = False
    res = testclient.get("/", status=500, expect_errors=True)
    res.mustcontain(no="LDAP authentication failed with user")



import datetime

import time_machine

from canaille.core.models import OTP_VALIDITY


def test_password_reset(testclient, user, backend):
    assert not backend.check_user_password(user, "foobarbaz")[0]
    token = user.generate_url_safe_token()
    backend.save(user)

    res = testclient.get("/reset/user/" + token, status=200)

    res.form["password"] = "foobarbaz"
    res.form["confirmation"] = "foobar"
    res = res.form.submit()
    res.mustcontain("Password and confirmation do not match.")
    res.mustcontain('data-percent="50"')

    res.form["password"] = "123"
    res.form["confirmation"] = "123"
    res = res.form.submit()
    res.mustcontain("Field must be at least 8 characters long.")

    res.form["password"] = "foobarbaz"
    res.form["confirmation"] = "foobarbaz"
    res = res.form.submit()
    assert ("success", "Your password has been updated successfully") in res.flashes
    assert res.location == "/profile/user"

    backend.reload(user)
    assert backend.check_user_password(user, "foobarbaz")[0]
    assert not user.one_time_password
    assert not user.one_time_password_emission_date

    res = testclient.get("/reset/user/" + token)
    assert (
        "error",
        "The password reset link that brought you here was invalid.",
    ) in res.flashes


def test_password_reset_expired_token(testclient, user, backend):
    with time_machine.travel("2020-01-01 01:00:00+00:00", tick=False) as traveller:
        token = user.generate_url_safe_token()
        backend.save(user)

        res = testclient.get("/reset/user/" + token, status=200)

        traveller.shift(datetime.timedelta(seconds=OTP_VALIDITY))

        res = testclient.get("/reset/user/" + token)
        assert (
            "error",
            "The password reset link that brought you here was invalid.",
        ) in res.flashes


def test_password_reset_multiple_emails(testclient, user, backend):
    user.emails = ["foo@bar.test", "foo@baz.test"]
    backend.save(user)

    assert not backend.check_user_password(user, "foobarbaz")[0]
    token = user.generate_url_safe_token()
    backend.save(user)

    res = testclient.get("/reset/user/" + token, status=200)

    res.form["password"] = "foobarbaz"
    res.form["confirmation"] = "foobarbaz"
    res = res.form.submit()
    assert ("success", "Your password has been updated successfully") in res.flashes

    backend.reload(user)
    assert backend.check_user_password(user, "foobarbaz")[0]

    res = testclient.get("/reset/user/" + token)
    assert (
        "error",
        "The password reset link that brought you here was invalid.",
    ) in res.flashes


def test_password_reset_bad_link(testclient, user):
    res = testclient.get("/reset/user/foobarbaz")
    assert (
        "error",
        "The password reset link that brought you here was invalid.",
    ) in res.flashes


def test_password_reset_bad_password(testclient, user, backend):
    token = user.generate_url_safe_token()
    backend.save(user)

    res = testclient.get("/reset/user/" + token, status=200)

    res.form["password"] = "foobarbaz"
    res.form["confirmation"] = "typo"
    res = res.form.submit(status=200)

    assert backend.check_user_password(user, "correct horse battery staple")[0]


def test_unavailable_if_no_smtp(testclient, user):
    res = testclient.get("/login")

    res.form["login"] = "user"
    res = res.form.submit()
    res = res.follow()
    res.mustcontain("Forgotten password")

    testclient.get("/reset", status=200)

    testclient.app.config["CANAILLE"]["SMTP"] = None

    res = testclient.get("/login")

    res.form["login"] = "user"
    res = res.form.submit()
    res = res.follow()
    res.mustcontain(no="Forgotten password")

    testclient.get("/reset", status=500, expect_errors=True)



import datetime
import os
from unittest import mock

import pytest
from babel.messages.frontend import compile_catalog
from flask_webtest import TestApp
from jinja2 import FileSystemBytecodeCache
from jinja2 import StrictUndefined
from pytest_lazy_fixtures import lf
from werkzeug.security import gen_salt

from canaille import create_app
from canaille.app import models
from canaille.app.session import UserSession
from canaille.backends import available_backends


@pytest.fixture(autouse=True, scope="session")
def babel_catalogs():
    cmd = compile_catalog()
    cmd.directory = os.path.dirname(__file__) + "/../canaille/translations"
    cmd.quiet = True
    cmd.statistics = True
    cmd.finalize_options()
    cmd.run()


pytest_plugins = [
    f"tests.backends.{backend}.fixtures"
    for backend in available_backends()
    if os.path.exists(os.path.join("tests", "backends", backend, "fixtures.py"))
]


def pytest_addoption(parser):
    parser.addoption(
        "--backend", action="append", default=[], help="the backends to test"
    )


def pytest_generate_tests(metafunc):
    backends = available_backends()
    if metafunc.config.getoption("backend"):  # pragma: no cover
        backends &= set(metafunc.config.getoption("backend"))

    # sort so the backend list is deterministic. This makes pytest-xdist discovery possible
    backends = sorted(backends)

    # tests in tests.backends.BACKENDNAME should only run one backend
    module_name_parts = metafunc.module.__name__.split(".")
    if module_name_parts[0:2] == ["tests", "backends"] and len(module_name_parts) > 3:
        backend = module_name_parts[2]
        if backend not in backends:  # pragma: no cover
            pytest.skip()
        elif "backend" in metafunc.fixturenames:
            metafunc.parametrize("backend", [lf(f"{backend}_backend")])
            return

    elif "backend" in metafunc.fixturenames:
        metafunc.parametrize(
            "backend",
            [lf(f"{backend}_backend") for backend in backends],
        )


@pytest.fixture
def configuration(smtpd):
    smtpd.config.use_starttls = True
    conf = {
        "SECRET_KEY": gen_salt(24),
        "SERVER_NAME": "canaille.test",
        "PREFERRED_URL_SCHEME": "http",
        "TRUSTED_HOSTS": [".canaille.test", "localhost", ".foobar.test"],
        "TESTING": True,
        "CANAILLE": {
            "JAVASCRIPT": False,
            "TIMEZONE": "UTC",
            "ACL": {
                "DEFAULT": {
                    "READ": ["user_name", "groups"],
                    "PERMISSIONS": ["edit_self", "use_oidc"],
                    "WRITE": [
                        "emails",
                        "given_name",
                        "photo",
                        "family_name",
                        "display_name",
                        "password",
                        "phone_numbers",
                        "formatted_address",
                        "street",
                        "postal_code",
                        "locality",
                        "region",
                        "employee_number",
                        "department",
                        "preferred_language",
                        "title",
                        "organization",
                        "lock_date",
                    ],
                },
                "ADMIN": {
                    "FILTER": [{"user_name": "admin"}, {"family_name": "admin"}],
                    "PERMISSIONS": [
                        "manage_users",
                        "manage_groups",
                        "manage_oidc",
                        "delete_account",
                        "impersonate_users",
                    ],
                    "WRITE": [
                        "groups",
                        "lock_date",
                    ],
                },
                "MODERATOR": {
                    "FILTER": [
                        {"user_name": "moderator"},
                        {"family_name": "moderator"},
                    ],
                    "PERMISSIONS": ["manage_users", "manage_groups", "delete_account"],
                    "WRITE": [
                        "groups",
                    ],
                },
            },
            "SMTP": {
                "HOST": smtpd.hostname,
                "PORT": smtpd.port,
                "TLS": smtpd.config.use_starttls,
                "SSL": smtpd.config.use_ssl,
                "LOGIN": smtpd.config.login_username,
                "PASSWORD": smtpd.config.login_password,
                "FROM_ADDR": "admin@mydomain.test",
            },
            "SMPP": {
                "HOST": "localhost",
                "PORT": 2775,
                "LOGIN": "user",
                "PASSWORD": "user",
            },
            "LOGGING": {
                "version": 1,
                "formatters": {
                    "default": {
                        "format": "[%(asctime)s] - %(levelname)s in %(module)s: %(message)s",
                    },
                    "wsgi": {
                        "format": "[%(asctime)s] - %(ip)s - %(levelname)s in %(module)s: %(message)s",
                    },
                },
                "handlers": {
                    "default": {
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stdout",
                        "formatter": "default",
                    },
                    "wsgi": {
                        "class": "logging.StreamHandler",
                        "stream": "ext://sys.stdout",
                        "formatter": "wsgi",
                    },
                },
                "loggers": {
                    "faker": {"level": "WARNING"},
                    "httpx": {"level": "WARNING"},
                    "httpcore": {"level": "WARNING"},
                    "canaille": {"level": "DEBUG", "handlers": ["wsgi"]},
                },
                "disable_existing_loggers": False,
            },
            "ADMIN_EMAIL": "admin_default_mail@mydomain.test",
            "PASSWORD_COMPROMISSION_CHECK_API_URL": "https://api.pwnedpasswords.test/range/",
        },
    }
    return conf


@pytest.fixture(autouse=True)
def email_validator():
    import email_validator

    email_validator.TEST_ENVIRONMENT = True


@pytest.fixture(scope="session")
def jinja_cache_directory(tmp_path_factory):
    return tmp_path_factory.mktemp("cache")


@pytest.fixture
def app(configuration, backend, jinja_cache_directory):
    app = create_app(configuration, backend=backend)
    # caches the Jinja compiled files for faster unit test execution
    app.jinja_env.bytecode_cache = FileSystemBytecodeCache(jinja_cache_directory)

    with app.app_context():
        backend.install(app)

    with app.test_request_context():
        yield app


@pytest.fixture
def testclient(app):
    app.jinja_env.undefined = StrictUndefined
    yield TestApp(app)


@pytest.fixture
def cli_runner(app):
    return app.test_cli_runner(catch_exceptions=False)


@pytest.fixture
def user(app, backend):
    u = models.User(
        formatted_name="John (johnny) Doe",
        given_name="John",
        family_name="Doe",
        user_name="user",
        emails=["john@doe.test"],
        password="correct horse battery staple",
        display_name="Johnny",
        preferred_language="en",
        phone_numbers=["555-000-000"],
        profile_url="https://john.test",
        formatted_address="1234, some street, 6789 some city, some state",
        street="1234, some street",
        locality="some city",
        postal_code="6789",
        region="some state",
        secret_token="fefe9b10",
        hotp_counter=1,
        last_otp_login=datetime.datetime(2020, 1, 1, tzinfo=datetime.timezone.utc),
    )
    backend.save(u)
    yield u
    backend.delete(u)


@pytest.fixture
def admin(app, backend):
    u = models.User(
        formatted_name="Jane Doe",
        family_name="Doe",
        user_name="admin",
        emails=["jane@doe.test"],
        password="admin",
    )
    backend.save(u)
    yield u
    backend.delete(u)


@pytest.fixture
def moderator(app, backend):
    u = models.User(
        formatted_name="Jack Doe",
        family_name="Doe",
        user_name="moderator",
        emails=["jack@doe.test"],
        password="moderator",
    )
    backend.save(u)
    yield u
    backend.delete(u)


@pytest.fixture
def logged_user(user, testclient):
    with testclient.session_transaction() as sess:
        sess["sessions"] = [
            UserSession(
                user=user,
                last_login_datetime=datetime.datetime.now(datetime.timezone.utc),
            ).serialize()
        ]
    return user


@pytest.fixture
def logged_admin(admin, testclient):
    with testclient.session_transaction() as sess:
        sess["sessions"] = [
            UserSession(
                user=admin,
                last_login_datetime=datetime.datetime.now(datetime.timezone.utc),
            ).serialize()
        ]
    return admin


@pytest.fixture
def logged_moderator(moderator, testclient):
    with testclient.session_transaction() as sess:
        sess["sessions"] = [
            UserSession(
                user=moderator,
                last_login_datetime=datetime.datetime.now(datetime.timezone.utc),
            ).serialize()
        ]
    return moderator


@pytest.fixture
def foo_group(app, user, backend):
    group = models.Group(
        members=[user],
        display_name="foo",
    )
    backend.save(group)
    backend.reload(user)
    yield group
    backend.delete(group)


@pytest.fixture
def bar_group(app, admin, backend):
    group = models.Group(
        members=[admin],
        display_name="bar",
    )
    backend.save(group)
    backend.reload(admin)
    yield group
    backend.delete(group)


@pytest.fixture
def jpeg_photo():
    with open("tests/fixtures/minimal.jpg", "rb") as fd:
        return fd.read()


@pytest.fixture(autouse=True)
def smpp_client():
    client = mock.Mock()
    client.__enter__ = client.Mock()
    client.__exit__ = client.Mock()
    client.__enter__.return_value = client
    with mock.patch("smpplib.client.Client", return_value=client):
        yield client



import pytest
from flask_webtest import TestApp

from canaille import create_app
from canaille.app.configuration import settings_factory
from canaille.app.installation import InstallationException
from canaille.backends.ldap.backend import LDAPBackend
from canaille.backends.ldap.ldapobject import LDAPObject
from canaille.commands import cli

from . import CustomSlapdObject


@pytest.fixture
def configuration(ldap_configuration):
    yield ldap_configuration


@pytest.fixture
def slapd_server():
    slapd = CustomSlapdObject()
    try:
        slapd.start()
        slapd.init_tree()
        for ldif in (
            "dev/ldif/memberof-config.ldif",
            "dev/ldif/bootstrap-users-tree.ldif",
        ):
            slapd.ldapadd(None, ["-f", ldif])

        yield slapd
    finally:
        slapd.stop()


def test_setup_ldap_tree(slapd_server, configuration):
    output = slapd_server.slapcat().stdout.decode("utf-8")
    assert "dn: ou=tokens,ou=oauth,dc=example,dc=org" not in output
    testclient = TestApp(create_app(configuration))
    cli_runner = testclient.app.test_cli_runner(catch_exceptions=False)

    res = cli_runner.invoke(cli, ["install"])
    assert res.exit_code == 0, res.stdout

    output = slapd_server.slapcat().stdout.decode("utf-8")
    assert "dn: ou=tokens,ou=oauth,dc=example,dc=org" in output


def test_install_schemas(configuration, slapd_server):
    configuration["CANAILLE_LDAP"]["ROOT_DN"] = slapd_server.suffix
    configuration["CANAILLE_LDAP"]["URI"] = slapd_server.ldap_uri
    configuration["CANAILLE_LDAP"]["BIND_DN"] = slapd_server.root_dn
    configuration["CANAILLE_LDAP"]["BIND_PW"] = slapd_server.root_pw
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()

    with LDAPBackend(config_dict).session():
        assert "oauthClient" not in LDAPObject.ldap_object_classes(force=True)

    LDAPBackend.setup_schemas(config_dict)

    with LDAPBackend(config_dict).session():
        assert "oauthClient" in LDAPObject.ldap_object_classes(force=True)


def test_install_schemas_twice(configuration, slapd_server):
    configuration["CANAILLE_LDAP"]["ROOT_DN"] = slapd_server.suffix
    configuration["CANAILLE_LDAP"]["URI"] = slapd_server.ldap_uri
    configuration["CANAILLE_LDAP"]["BIND_DN"] = slapd_server.root_dn
    configuration["CANAILLE_LDAP"]["BIND_PW"] = slapd_server.root_pw
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()

    with LDAPBackend(config_dict).session():
        assert "oauthClient" not in LDAPObject.ldap_object_classes(force=True)

    LDAPBackend.setup_schemas(config_dict)

    with LDAPBackend(config_dict).session():
        assert "oauthClient" in LDAPObject.ldap_object_classes(force=True)

    LDAPBackend.setup_schemas(config_dict)


admin_ldif = """
dn: uid=admin,ou=users,dc=example,dc=org
objectclass: top
objectclass: inetOrgPerson
cn: Jane Doe
givenName: Jane
sn: Doe
uid: admin
mail: admin@example.org
userPassword: {SSHA}7zQVLckaEc6cJEsS0ylVipvb2PAR/4tS
displayName: Jane.D
"""


def test_install_no_permissions_to_install_schemas(configuration, slapd_server):
    slapd_server.slapadd(admin_ldif)

    configuration["CANAILLE_LDAP"]["ROOT_DN"] = slapd_server.suffix
    configuration["CANAILLE_LDAP"]["URI"] = slapd_server.ldap_uri
    configuration["CANAILLE_LDAP"]["BIND_DN"] = "uid=admin,ou=users,dc=example,dc=org"
    configuration["CANAILLE_LDAP"]["BIND_PW"] = "admin"
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()

    with LDAPBackend(config_dict).session():
        assert "oauthClient" not in LDAPObject.ldap_object_classes(force=True)

        with pytest.raises(InstallationException):
            LDAPBackend.setup_schemas(config_dict)

        assert "oauthClient" not in LDAPObject.ldap_object_classes(force=True)


def test_install_schemas_command(configuration, slapd_server):
    configuration["CANAILLE_LDAP"]["ROOT_DN"] = slapd_server.suffix
    configuration["CANAILLE_LDAP"]["URI"] = slapd_server.ldap_uri
    configuration["CANAILLE_LDAP"]["BIND_DN"] = slapd_server.root_dn
    configuration["CANAILLE_LDAP"]["BIND_PW"] = slapd_server.root_pw
    config_obj = settings_factory(configuration)
    config_dict = config_obj.model_dump()

    with LDAPBackend(config_dict).session():
        assert "oauthClient" not in LDAPObject.ldap_object_classes(force=True)

    testclient = TestApp(create_app(configuration))
    cli_runner = testclient.app.test_cli_runner(catch_exceptions=False)

    res = cli_runner.invoke(cli, ["install"])
    assert res.exit_code == 0, res.stdout

    with LDAPBackend(config_dict).session():
        assert "oauthClient" in LDAPObject.ldap_object_classes(force=True)



import logging
from urllib.parse import parse_qs
from urllib.parse import urlsplit

from canaille.app import models

from . import client_credentials


def test_no_logged_no_access(testclient):
    testclient.get("/consent", status=403)


def test_revokation(testclient, client, consent, logged_user, token, backend, caplog):
    res = testclient.get("/consent", status=200)
    res.mustcontain(client.client_name)
    res.mustcontain("Revoke access")
    res.mustcontain(no="Restore access")
    assert not consent.revoked
    assert not token.revoked

    res = testclient.get(f"/consent/revoke/{consent.consent_id}", status=302)
    assert ("success", "The access has been revoked") in res.flashes
    assert (
        "canaille",
        logging.SECURITY,
        f"Consent revoked for {logged_user.user_name} in client {client.client_name}",
    ) in caplog.record_tuples
    res = res.follow(status=200)
    res.mustcontain(no="Revoke access")
    res.mustcontain("Restore access")

    backend.reload(consent)
    assert consent.revoked
    backend.reload(token)
    assert token.revoked


def test_revokation_already_revoked(testclient, client, consent, logged_user, backend):
    consent.revoke()

    assert consent.revoked

    res = testclient.get(f"/consent/revoke/{consent.consent_id}", status=302)
    assert ("error", "The access is already revoked") in res.flashes
    res = res.follow(status=200)

    backend.reload(consent)
    assert consent.revoked


def test_restoration(testclient, client, consent, logged_user, token, backend):
    consent.revoke()

    assert consent.revoked
    backend.reload(token)
    assert token.revoked

    res = testclient.get(f"/consent/restore/{consent.consent_id}", status=302)
    assert ("success", "The access has been restored") in res.flashes
    res = res.follow(status=200)

    backend.reload(consent)
    assert not consent.revoked
    backend.reload(token)
    assert token.revoked


def test_restoration_already_restored(testclient, client, consent, logged_user, token):
    assert not consent.revoked

    res = testclient.get(f"/consent/restore/{consent.consent_id}", status=302)
    assert ("error", "The access is not revoked") in res.flashes
    res = res.follow(status=200)


def test_invalid_consent_revokation(testclient, client, logged_user):
    res = testclient.get("/consent/revoke/invalid", status=302)
    assert ("success", "The access has been revoked") not in res.flashes
    assert ("error", "Could not revoke this access") in res.flashes


def test_someone_else_consent_revokation(testclient, client, consent, logged_moderator):
    res = testclient.get(f"/consent/revoke/{consent.consent_id}", status=302)
    assert ("success", "The access has been revoked") not in res.flashes
    assert ("error", "Could not revoke this access") in res.flashes


def test_invalid_consent_restoration(testclient, client, logged_user):
    res = testclient.get("/consent/restore/invalid", status=302)
    assert ("success", "The access has been restored") not in res.flashes
    assert ("error", "Could not restore this access") in res.flashes


def test_someone_else_consent_restoration(
    testclient, client, consent, logged_moderator
):
    res = testclient.get(f"/consent/restore/{consent.consent_id}", status=302)
    assert ("success", "The access has been restore") not in res.flashes
    assert ("error", "Could not restore this access") in res.flashes


def test_oidc_authorization_after_revokation(
    testclient, logged_user, client, consent, backend
):
    consent.revoke()

    assert consent.revoked

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
        status=200,
    )

    res = res.form.submit(name="answer", value="accept", status=302)

    consents = backend.query(models.Consent, client=client, subject=logged_user)
    backend.reload(consent)
    assert consents[0] == consent
    assert not consent.revoked

    params = parse_qs(urlsplit(res.location).query)
    code = params["code"][0]
    res = testclient.post(
        "/oauth/token",
        params=dict(
            grant_type="authorization_code",
            code=code,
            scope="openid profile",
            redirect_uri=client.redirect_uris[0],
        ),
        headers={"Authorization": f"Basic {client_credentials(client)}"},
        status=200,
    )

    access_token = res.json["access_token"]
    token = backend.get(models.Token, access_token=access_token)
    assert token.client == client
    assert token.subject == logged_user


def test_trusted_client_appears_in_consent_list(
    testclient, client, logged_user, backend
):
    assert not client.trusted
    res = testclient.get("/consent/trusted-applications")
    res.mustcontain(no=client.client_name)

    client.client_uri = "https://client.trusted.test"
    backend.save(client)

    res = testclient.get("/consent/trusted-applications")
    res.mustcontain(client.client_name)


def test_revoke_trusted_client(testclient, client, logged_user, token, backend):
    client.client_uri = "https://client.trusted.test"
    backend.save(client)
    assert not backend.get(models.Consent)
    assert not token.revoked

    res = testclient.get(f"/consent/revoke-trusted/{client.client_id}", status=302)
    assert ("success", "The access has been revoked") in res.flashes

    consent = backend.get(models.Consent)
    assert consent.client == client
    assert consent.subject == logged_user
    assert consent.scope == ["openid", "email", "profile", "groups", "address", "phone"]
    assert not consent.issue_date
    backend.reload(token)
    assert token.revoked

    res = testclient.get(f"/consent/restore/{consent.consent_id}", status=302)
    assert ("success", "The access has been restored") in res.flashes

    backend.reload(consent)
    assert not consent.revoked
    assert consent.issue_date
    backend.reload(token)
    assert token.revoked

    res = testclient.get(f"/consent/revoke/{consent.consent_id}", status=302)
    assert ("success", "The access has been revoked") in res.flashes
    backend.reload(consent)
    assert consent.revoked
    assert consent.issue_date


def test_revoke_invalid_trusted_client(testclient, logged_user):
    res = testclient.get("/consent/revoke-trusted/invalid", status=302)
    assert ("error", "Could not revoke this access") in res.flashes


def test_revoke_trusted_client_with_manual_consent(
    testclient, logged_user, client, consent, backend
):
    client.client_uri = "https://client.trusted.test"
    backend.save(client)
    res = testclient.get(f"/consent/revoke-trusted/{client.client_id}", status=302)
    res = res.follow()
    assert ("success", "The access has been revoked") in res.flashes


def test_revoke_trusted_client_with_manual_revokation(
    testclient, logged_user, client, consent, backend
):
    client.client_uri = "https://client.trusted.test"
    backend.save(client)
    consent.revoke()
    backend.save(consent)

    res = testclient.get(f"/consent/revoke-trusted/{client.client_id}", status=302)
    res = res.follow()
    assert ("error", "The access is already revoked") in res.flashes


def test_tos_policy(testclient, logged_user, client, trusted_client, backend, caplog):
    """Tos and policy links must appear in the consent page."""
    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="code",
            client_id=client.client_id,
            scope="openid profile email groups address phone",
            nonce="somenonce",
            tos_uri="https://client.test/tos",
            policy_uri="https://client.test/policy",
            redirect_uri="https://client.test/redirect1",
        ),
        status=200,
    )
    res.mustcontain("https://client.test/tos")
    res.mustcontain("https://client.test/policy")



import json

import click
from flask.cli import with_appcontext
from joserfc.jwk import ECKey
from joserfc.jwk import OctKey
from joserfc.jwk import OKPKey
from joserfc.jwk import RSAKey

from canaille.app import models
from canaille.app.commands import with_backendcontext
from canaille.backends import Backend


@click.command()
@with_appcontext
@with_backendcontext
def clean():
    """Remove expired tokens and authorization codes."""
    for t in Backend.instance.query(models.Token):
        if t.is_expired():
            Backend.instance.delete(t)

    for a in Backend.instance.query(models.AuthorizationCode):
        if a.is_expired():
            Backend.instance.delete(a)


@click.group()
def jwk():
    """JSON Web Key management."""
    pass


@jwk.group()
def create():
    """JSON Web Key creation."""
    pass


@create.command()
@click.option("--size", default=2048, type=int, help="The key size")
def rsa(size: int):
    """Create a RSA JSON Web Key."""
    key = RSAKey.generate_key(size, auto_kid=True)
    click.echo(json.dumps(key.as_dict()))


@create.command()
@click.option("--size", default=256, type=int, help="The key size")
def oct(size: int):
    """Create a Oct JSON Web Key."""
    key = OctKey.generate_key(size, auto_kid=True)
    click.echo(json.dumps(key.as_dict()))


@create.command()
@click.option("--crv", default="P-256", type=str, help="The key CRV")
def ec(crv: str):
    """Create a EC JSON Web Key."""
    key = ECKey.generate_key(crv, auto_kid=True)
    click.echo(json.dumps(key.as_dict()))


@create.command()
@click.option("--crv", default="Ed25519", type=str, help="The key CRV")
def okp(crv: str):
    """Create a OKP JSON Web Key."""
    key = OKPKey.generate_key(crv, auto_kid=True)
    click.echo(json.dumps(key.as_dict()))


def register(cli):
    cli.add_command(clean)
    cli.add_command(jwk)



import datetime
import secrets
from typing import Annotated
from typing import ClassVar

from blinker import signal
from flask import current_app
from pydantic import TypeAdapter

from canaille.app import build_hash
from canaille.backends.models import Model
from canaille.core.configuration import Permission

OTP_DIGITS = 6
OTP_VALIDITY = 600
SEND_NEW_OTP_DELAY = 10

PASSWORD_MIN_DELAY = 2
PASSWORD_MAX_DELAY = 600
PASSWORD_FAILURE_COUNT_INTERVAL = 600

LINK_TOKEN_LENGTH = 16


class User(Model):
    """User model.

    Based on the `SCIM User schema
    <https://datatracker.ietf.org/doc/html/rfc7643#section-4.1>`_,
    `Entreprise User Schema Extension
    <https://datatracker.ietf.org/doc/html/rfc7643#section-4.3>`_
    and `SCIM Password Management Extension
    <https://datatracker.ietf.org/doc/html/draft-hunt-scim-password-mgmt-00.html>`_
    draft.
    Attribute description is based on SCIM and put there for
    information purpose. The description may not fit the current
    implementation in Canaille.
    """

    identifier_attribute: ClassVar[str] = "user_name"

    user_name: str
    """A service provider's unique identifier for the user, typically used by
    the user to directly authenticate to the service provider.

    Often displayed to the user as their unique identifier within the
    system (as opposed to "id" or "externalId", which are generally
    opaque and not user-friendly identifiers).  Each User MUST include a
    non-empty userName value.  This identifier MUST be unique across the
    service provider's entire set of Users.  This attribute is REQUIRED
    and is case insensitive.
    """

    password_failure_timestamps: list[datetime.datetime] = []
    """This attribute stores the timestamps of the user's failed
    authentications.

    It's currently used by the intruder lockout delay system.
    """

    password: str | None = None
    """
    This attribute is intended to be used as a means to set, replace,
    or compare (i.e., filter for equality) a password.  The cleartext
    value or the hashed value of a password SHALL NOT be returnable by
    a service provider.  If a service provider holds the value
    locally, the value SHOULD be hashed.  When a password is set or
    changed by the client, the cleartext password SHOULD be processed
    by the service provider as follows:

    *  Prepare the cleartext value for international language
       comparison.  See Section 7.8 of [RFC7644].

    *  Validate the value against server password policy.  Note: The
       definition and enforcement of password policy are beyond the
       scope of this document.

    *  Ensure that the value is encrypted (e.g., hashed).  See
       Section 9.2 for acceptable hashing and encryption handling when
       storing or persisting for provisioning workflow reasons.

    A service provider that immediately passes the cleartext value on
    to another system or programming interface MUST pass the value
    directly over a secured connection (e.g., Transport Layer Security
    (TLS)).  If the value needs to be temporarily persisted for a
    period of time (e.g., because of a workflow) before provisioning,
    then the value MUST be protected by some method, such as
    encryption.

    Testing for an equality match MAY be supported if there is an
    existing stored hashed value.  When testing for equality, the
    service provider:

    *  Prepares the filter value for international language
       comparison.  See Section 7.8 of [RFC7644].

    *  Generates the salted hash of the filter value and tests for a
       match with the locally held value.

    The mutability of the password attribute is "writeOnly",
    indicating that the value MUST NOT be returned by a service
    provider in any form (the attribute characteristic "returned" is
    "never").
    """

    password_last_update: datetime.datetime | None = None
    """Specifies the last time the user password was changed.
    By default, the date of creation of the password is retained.
    """

    preferred_language: str | None = None
    """Indicates the user's preferred written or spoken languages and is
    generally used for selecting a localized user interface.

    The value indicates the set of natural languages that are preferred.
    The format of the value is the same as the HTTP Accept-Language
    header field (not including "Accept-Language:") and is specified in
    Section 5.3.5 of [RFC7231].  The intent of this value is to enable
    cloud applications to perform matching of language tags [RFC4647] to
    the user's language preferences, regardless of what may be indicated
    by a user agent (which might be shared), or in an interaction that
    does not involve a user (such as in a delegated OAuth 2.0 [RFC6749]
    style interaction) where normal HTTP Accept-Language header
    negotiation cannot take place.
    """

    family_name: str | None = None
    """The family name of the User, or last name in most Western languages
    (e.g., "Jensen" given the full name "Ms. Barbara Jane Jensen, III")."""

    given_name: str | None = None
    """The given name of the User, or first name in most Western languages
    (e.g., "Barbara" given the full name "Ms. Barbara Jane Jensen, III")."""

    formatted_name: str | None = None
    """The full name, including all middle names, titles, and suffixes as
    appropriate, formatted for display (e.g., "Ms. Barbara Jane Jensen,
    III")."""

    display_name: str | None = None
    """The name of the user, suitable for display to end-users.

    Each user returned MAY include a non-empty displayName value.  The
    name SHOULD be the full name of the User being described, if known
    (e.g., "Babs Jensen" or "Ms. Barbara J Jensen, III") but MAY be a
    username or handle, if that is all that is available (e.g.,
    "bjensen").  The value provided SHOULD be the primary textual label
    by which this User is normally displayed by the service provider
    when presenting it to end-users.
    """

    emails: list[str] = []
    """Email addresses for the User.

    The value SHOULD be specified according to [RFC5321].  Service
    providers SHOULD canonicalize the value according to [RFC5321],
    e.g., "bjensen@example.com" instead of "bjensen@EXAMPLE.COM".  The
    "display" sub-attribute MAY be used to return the canonicalized
    representation of the email value. The "type" sub-attribute is used
    to provide a classification meaningful to the (human) user.  The
    user interface should encourage the use of basic values of "work",
    "home", and "other" and MAY allow additional type values to be used
    at the discretion of SCIM clients.
    """

    phone_numbers: list[str] = []
    """Phone numbers for the user.

    The value SHOULD be specified according to the format defined in
    [RFC3966], e.g., 'tel:+1-201-555-0123'.  Service providers SHOULD
    canonicalize the value according to [RFC3966] format, when
    appropriate.  The "display" sub-attribute MAY be used to return the
    canonicalized representation of the phone number value.  The sub-
    attribute "type" often has typical values of "work", "home",
    "mobile", "fax", "pager", and "other" and MAY allow more types to be
    defined by the SCIM clients.
    """

    formatted_address: str | None = None
    """The full mailing address, formatted for display or use with a mailing
    label.

    This attribute MAY contain newlines.
    """

    street: str | None = None
    """The full street address component, which may include house number,
    street name, P.O.

    box, and multi-line extended street address information.  This
    attribute MAY contain newlines.
    """

    postal_code: str | None = None
    """The zip code or postal code component."""

    locality: str | None = None
    """The city or locality component."""

    region: str | None = None
    """The state or region component."""

    photo: str | None = None
    """A URI that is a uniform resource locator (as defined in Section 1.1.3 of
    [RFC3986]) that points to a resource location representing the user's
    image.

    The resource MUST be a file (e.g.,
    a GIF, JPEG, or PNG image file) rather than a web page containing
    an image.  Service providers MAY return the same image in
    different sizes, although it is recognized that no standard for
    describing images of various sizes currently exists.  Note that
    this attribute SHOULD NOT be used to send down arbitrary photos
    taken by this user; instead, profile photos of the user that are
    suitable for display when describing the user should be sent.
    Instead of the standard canonical values for type, this attribute
    defines the following canonical values to represent popular photo
    sizes: "photo" and "thumbnail".
    """

    profile_url: str | None = None
    """A URI that is a uniform resource locator (as defined in Section 1.1.3 of
    [RFC3986]) and that points to a location representing the user's online
    profile (e.g., a web page).

    URIs are canonicalized per Section 6.2 of [RFC3986].
    """

    title: str | None = None
    """The user's title, such as "Vice President"."""

    organization: str | None = None
    """Identifies the name of an organization."""

    employee_number: str | None = None
    """A string identifier, typically numeric or alphanumeric, assigned to a
    person, typically based on order of hire or association with an
    organization."""

    department: str | None = None
    """Identifies the name of a department."""

    groups: list[Annotated["Group", {"backref": "members"}]] = []
    """A list of groups to which the user belongs, either through direct
    membership, through nested groups, or dynamically calculated.

    The values are meant to enable expression of common group-based or
    role-based access control models, although no explicit authorization
    model is defined.  It is intended that the semantics of group
    membership and any behavior or authorization granted as a result of
    membership are defined by the service provider.  The canonical types
    "direct" and "indirect" are defined to describe how the group
    membership was derived.  Direct group membership indicates that the
    user is directly associated with the group and SHOULD indicate that
    clients may modify membership through the "Group" resource. Indirect
    membership indicates that user membership is transitive or dynamic
    and implies that clients cannot modify indirect group membership
    through the "Group" resource but MAY modify direct group membership
    through the "Group" resource, which may influence indirect
    memberships.  If the SCIM service provider exposes a "Group"
    resource, the "value" sub-attribute MUST be the "id", and the "$ref"
    sub-attribute must be the URI of the corresponding "Group" resources
    to which the user belongs.  Since this attribute has a mutability of
    "readOnly", group membership changes MUST be applied via the "Group"
    Resource (Section 4.2).  This attribute has a mutability of
    "readOnly".
    """

    lock_date: datetime.datetime | None = None
    """A DateTime indicating when the resource was locked."""

    last_otp_login: datetime.datetime | None = None
    """A DateTime indicating when the user last logged in with a one-time passcode.
    This attribute is currently used to check whether the user has activated one-time passcode authentication or not."""

    secret_token: str | None = None
    """Unique token generated for each user, used for
    multi-factor authentication."""

    hotp_counter: int | None = None
    """HMAC-based One Time Password counter, used for
    multi-factor authentication."""

    one_time_password: str | None = None
    """One time password used for email or sms multi-factor authentication."""

    one_time_password_emission_date: datetime.datetime | None = None
    """A DateTime indicating when the user last emitted an email or sms one-time passcode."""

    _readable_fields: set[str] | None = None
    _writable_fields: set[str] | None = None
    _permissions: set[str] | None = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        signal("before_user_reload").connect(self.on_reload, sender=self)

    def has_password(self) -> bool:
        """Check whether a password has been set for the user."""
        return self.password is not None

    def can_read(self, field: str):
        return field in self.readable_fields | self.writable_fields

    @property
    def name(self):
        return self.display_name or self.formatted_name or self.user_name

    @property
    def preferred_email(self):
        return self.emails[0] if self.emails else None

    def __getattribute__(self, name):
        prefix = "can_"

        try:
            return super().__getattribute__(name)

        except AttributeError:
            if name.startswith(prefix) and name != "can_read":
                return self.can(name[len(prefix) :])
            raise

    def can(self, *permissions: Permission):
        """Whether or not the user has the :class:`~canaille.core.configuration.Permission` according to the :class:`configuration <canaille.core.configuration.ACLSettings>`."""
        if self._permissions is None:
            self._permissions = set()
            acls = current_app.config["CANAILLE"]["ACL"].values()
            for details in acls:
                if self.match_filter(details["FILTER"]):
                    self._permissions |= set(details["PERMISSIONS"])

        return set(permissions).issubset(self._permissions)

    @property
    def locked(self) -> bool:
        """Whether the user account has been locked or has expired."""
        return self.lock_date is not None and self.lock_date < datetime.datetime.now(
            datetime.timezone.utc
        )

    @classmethod
    def on_reload(cls, self, data):
        self._readable = None
        self._writable = None
        self._permissions = None

    @property
    def readable_fields(self):
        """The fields the user can read according to the :class:`configuration <canaille.core.configuration.ACLSettings>` configuration.

        This does not include the :attr:`writable
        <canaille.core.models.User.writable_fields>` fields.
        """
        if self._readable_fields is None:
            self._readable_fields = set()
            acls = current_app.config["CANAILLE"]["ACL"].values()
            for details in acls:
                if self.match_filter(details["FILTER"]):
                    self._readable_fields |= set(details["READ"])

        return self._readable_fields

    @property
    def writable_fields(self):
        """The fields the user can write according to the :class:`configuration <canaille.core.configuration.ACLSettings>`."""
        if self._writable_fields is None:
            self._writable_fields = set()
            acls = current_app.config["CANAILLE"]["ACL"].values()
            for details in acls:
                if self.match_filter(details["FILTER"]):
                    self._writable_fields |= set(details["WRITE"])
        return self._writable_fields

    def generate_sms_or_mail_otp(self, length=OTP_DIGITS):
        otp = string_code(secrets.randbelow(10**length), length)
        self.one_time_password = otp
        self.one_time_password_emission_date = datetime.datetime.now(
            datetime.timezone.utc
        )
        return otp

    def generate_url_safe_token(self):
        token = secrets.token_urlsafe(LINK_TOKEN_LENGTH)
        self.one_time_password = build_hash(token)
        self.one_time_password_emission_date = datetime.datetime.now(
            datetime.timezone.utc
        )
        return token

    def is_email_or_sms_otp_valid(self, user_otp):
        return user_otp == self.one_time_password and self.is_otp_still_valid()

    def is_otp_still_valid(self):
        return datetime.datetime.now(
            datetime.timezone.utc
        ) - self.one_time_password_emission_date < datetime.timedelta(
            seconds=OTP_VALIDITY
        )

    def can_send_new_otp(self):
        return self.one_time_password_emission_date is None or (
            datetime.datetime.now(datetime.timezone.utc)
            - self.one_time_password_emission_date
            >= datetime.timedelta(seconds=SEND_NEW_OTP_DELAY)
        )

    @property
    def next_otp_send_delay(self):
        if self.can_send_new_otp():
            return datetime.timedelta(seconds=0)

        now = datetime.datetime.now(datetime.timezone.utc)
        return (
            self.one_time_password_emission_date
            + datetime.timedelta(seconds=SEND_NEW_OTP_DELAY)
        ) - now

    def clear_otp(self):
        self.one_time_password_emission_date = None
        self.one_time_password = None

    def get_intruder_lockout_delay(self):
        if self.password_failure_timestamps:
            # discard old attempts
            self.password_failure_timestamps = [
                attempt
                for attempt in self.password_failure_timestamps
                if attempt
                > datetime.datetime.now(datetime.timezone.utc)
                - datetime.timedelta(seconds=PASSWORD_FAILURE_COUNT_INTERVAL)
            ]
        if not self.password_failure_timestamps:
            return 0
        failed_login_count = len(self.password_failure_timestamps)
        # delay is multiplied by 2 each failed attempt, starting at min delay, limited to max delay
        calculated_delay = min(
            PASSWORD_MIN_DELAY * 2 ** (failed_login_count - 1), PASSWORD_MAX_DELAY
        )
        time_since_last_failed_bind = (
            datetime.datetime.now(datetime.timezone.utc)
            - self.password_failure_timestamps[-1]
        ).total_seconds()
        return max(calculated_delay - time_since_last_failed_bind, 0)

    def has_expired_password(self):
        last_update = self.password_last_update or datetime.datetime.now(
            datetime.timezone.utc
        )
        if current_app.config["CANAILLE"]["PASSWORD_LIFETIME"] is None:
            password_expiration = None
        else:
            password_expiration = TypeAdapter(datetime.timedelta).validate_python(
                current_app.config["CANAILLE"]["PASSWORD_LIFETIME"]
            )

        return (
            password_expiration is not None
            and last_update + password_expiration
            < datetime.datetime.now(datetime.timezone.utc)
        )


class Group(Model):
    """Group model.

    Based on the `SCIM Group schema
    <https://datatracker.ietf.org/doc/html/rfc7643#section-4.2>`_.
    """

    identifier_attribute: ClassVar[str] = "display_name"

    display_name: str
    """A human-readable name for the Group.

    REQUIRED.
    """

    members: list[Annotated["User", {"backref": "groups"}]] = []
    """A list of members of the Group.

    While values MAY be added or removed, sub-attributes of members are
    "immutable".  The "value" sub-attribute contains the value of an
    "id" attribute of a SCIM resource, and the "$ref" sub-attribute must
    be the URI of a SCIM resource such as a "User", or a "Group".  The
    intention of the "Group" type is to allow the service provider to
    support nested groups.  Service providers MAY require clients to
    provide a non-empty value by setting the "required" attribute
    characteristic of a sub-attribute of the "members" attribute in the
    "Group" resource schema.
    """

    description: str | None = None


def string_code(code: int, digit: int) -> str:
    """Add leading 0 if the code length does not match the defined length.

    For instance, parameter ``digit=6``, but ``code=123``, this method would
    return ``000123``:

    .. code-block:: python

        >>> otp.string_code(123)
        '000123'
    """
    return f"{code:0{digit}}"



import datetime

from flask import current_app
from flask import g
from flask import request

DEFAULT_LANGUAGE_CODE = "en"

try:
    from flask_babel import Babel
    from flask_babel import get_locale
    from flask_babel import gettext
    from flask_babel import lazy_gettext

    babel = Babel()
except ImportError:

    def identity(string, *args, **kwargs):
        return string

    def get_locale():
        return "en_US"

    gettext = identity
    lazy_gettext = identity
    babel = None


def mock_babel(app):  # pragma: no cover
    """Jinja stubs when flask-babel is not installed."""
    app.jinja_env.filters.update(
        datetimeformat=identity,
        dateformat=identity,
        timeformat=identity,
        timedeltaformat=identity,
        numberformat=identity,
        decimalformat=identity,
        currencyformat=identity,
        percentformat=identity,
        scientificformat=identity,
    )
    app.jinja_env.add_extension("jinja2.ext.i18n")
    app.jinja_env.install_gettext_callables(
        gettext=identity,
        ngettext=identity,
        newstyle=True,
        pgettext=identity,
        npgettext=identity,
    )


def setup_i18n(app):
    @app.before_request
    def before_request():
        g.available_language_codes = available_language_codes()

    @app.context_processor
    def global_processor():
        return {
            "locale": get_locale(),
        }

    if not babel:  # pragma: no cover
        mock_babel(app)
        return

    babel.init_app(
        app, locale_selector=locale_selector, timezone_selector=timezone_selector
    )


def smart_language_match(accept_languages, available_languages):
    """Match accepted languages to available ones, handling regional variants.

    Improves upon Werkzeug's best_match by falling back to base language codes
    when exact regional variants aren't available. For example, if 'fr-FR' is
    requested but only 'fr' is available, returns 'fr' instead of falling back
    to a lower priority language like 'en'.

    :param accept_languages: LanguageAccept object with client's language preferences
    :param available_languages: List of available language codes
    :returns: Best matching language code from available_languages
    """
    exact_match = accept_languages.best_match(available_languages)

    if exact_match and exact_match != "en":
        return exact_match

    for lang_code, _priority in accept_languages:
        base_lang = lang_code.split("-")[0].split("_")[0]

        if base_lang in available_languages:
            return base_lang

        for available in available_languages:
            if (
                available.startswith(base_lang + "_")
                or available.split("_")[0] == base_lang
            ):
                return available

    return exact_match or DEFAULT_LANGUAGE_CODE


def locale_selector():
    available_language_codes = getattr(g, "available_language_codes", [])
    if (
        g.get("session")
        and g.session.user.preferred_language in available_language_codes
    ):
        return g.session.user.preferred_language

    if current_app.config["CANAILLE"]["LANGUAGE"]:
        return current_app.config["CANAILLE"]["LANGUAGE"]

    return smart_language_match(request.accept_languages, available_language_codes)


def timezone_selector():
    if not babel:  # pragma: no cover
        return datetime.timezone.utc

    import pytz
    from babel.dates import LOCALTZ

    try:
        return pytz.timezone(current_app.config["CANAILLE"]["TIMEZONE"])
    except pytz.exceptions.UnknownTimeZoneError:
        return LOCALTZ


def native_language_name_from_code(code):
    try:
        from gettext import translation

        import pycountry
    except ImportError:
        return code

    language = pycountry.languages.get(alpha_2=code[:2])
    if code == DEFAULT_LANGUAGE_CODE:
        return language.name

    translation = translation("iso639-3", pycountry.LOCALES_DIR, languages=[code])
    return translation.gettext(language.name)


def available_language_codes():
    return (
        [str(translation) for translation in babel.list_translations()]
        if babel
        else [DEFAULT_LANGUAGE_CODE]
    )


def reload_translations():
    if not babel:  # pragma: no cover
        return

    from flask_babel import refresh

    return refresh()



from canaille.commands import cli


def test_check_command(cli_runner):
    res = cli_runner.invoke(cli, ["config", "check"])
    assert res.exit_code == 0, res.stdout


def test_check_command_fail(testclient, cli_runner):
    testclient.app.config["CANAILLE"]["SMTP"]["HOST"] = "invalid-domain.com"

    res = cli_runner.invoke(cli, ["config", "check"])
    assert res.exit_code == 1, res.stdout


def test_check_command_no_smtp(testclient, cli_runner):
    testclient.app.config["CANAILLE"]["SMTP"] = None

    res = cli_runner.invoke(cli, ["config", "check"])
    assert res.exit_code == 0, res.stdout



from canaille.oidc.jwk import make_default_jwk


def test_deterministic_jwk_generation_with_seed():
    """Test that the same seed always generates the same JWK."""
    test_seed = "test-seed-123"

    jwk1 = make_default_jwk(test_seed).as_dict()
    jwk2 = make_default_jwk(test_seed).as_dict()
    jwk3 = make_default_jwk(test_seed).as_dict()

    assert jwk1 == jwk2
    assert jwk2 == jwk3
    assert jwk1 == jwk3

    assert "kid" in jwk1
    assert "kty" in jwk1
    assert jwk1["kty"] == "OKP"
    assert jwk1["crv"] == "Ed25519"

    jwk4 = make_default_jwk("different-seed-456").as_dict()

    assert jwk1 != jwk4

    jwk5 = make_default_jwk("different-seed-456").as_dict()
    assert jwk4 == jwk5


def test_random_jwk_generation_without_seed():
    """Test that without seed, JWKs are randomly generated."""
    jwk1 = make_default_jwk(None).as_dict()
    jwk2 = make_default_jwk(None).as_dict()

    assert jwk1 != jwk2

    assert "kty" in jwk1
    assert jwk1["kty"] == "OKP"
    assert jwk1["crv"] == "Ed25519"


def test_jwks(testclient, server_jwk, old_server_jwk):
    res = testclient.get("/oauth/jwks.json")
    assert res.json == {
        "keys": [
            server_jwk.as_dict(private=False),
            old_server_jwk.as_dict(private=False),
        ]
    }



def test_group_permissions_by_id(testclient, user, foo_group, backend):
    assert not user.can_manage_users

    testclient.app.config["CANAILLE"]["ACL"]["ADMIN"]["FILTER"] = {
        "groups": foo_group.id
    }
    backend.reload(user)

    assert user.can_manage_users


def test_group_permissions_by_display_name(testclient, user, foo_group, backend):
    assert not user.can_manage_users

    testclient.app.config["CANAILLE"]["ACL"]["ADMIN"]["FILTER"] = {
        "groups": foo_group.display_name
    }
    backend.reload(user)

    assert user.can_manage_users


def test_invalid_group_permission(testclient, user, foo_group, backend):
    assert not user.can_manage_users

    testclient.app.config["CANAILLE"]["ACL"]["ADMIN"]["FILTER"] = {"groups": "invalid"}
    backend.reload(user)

    assert not user.can_manage_users



import datetime

from flask import g

from canaille.app import models
from canaille.core.endpoints.account import RegistrationPayload


def test_invitation(testclient, logged_admin, foo_group, smtpd, backend):
    assert backend.get(models.User, user_name="someone") is None

    res = testclient.get("/invite", status=200)

    res.form["user_name"] = "someone"
    res.form["user_name_editable"] = False
    res.form["email"] = "someone@domain.test"
    res.form["groups"] = [foo_group.id]
    res = res.form.submit(name="action", value="send", status=200)
    assert len(smtpd.messages) == 1

    url = res.pyquery(".copy-text")[0].value

    # logout
    g.session = None
    with testclient.session_transaction() as sess:
        del sess["sessions"]

    testclient.get("/logout")
    res = testclient.get(url, status=200)

    assert "readonly" in res.form["user_name"].attrs
    assert "readonly" in res.form["emails-0"].attrs
    assert "readonly" in res.form["groups"].attrs

    assert res.form["user_name"].value == "someone"
    assert res.form["emails-0"].value == "someone@domain.test"
    assert res.form["groups"].value == [foo_group.id]

    res.form["password1"] = "i'm a little pea"
    res.form["password2"] = "i'm a little pea"
    res.form["given_name"] = "George"
    res.form["family_name"] = "Abitbol"

    res = res.form.submit(status=302)

    assert ("success", "Your account has been created successfully.") in res.flashes
    res = res.follow(status=200)

    user = backend.get(models.User, user_name="someone")
    backend.reload(foo_group)
    assert backend.check_user_password(user, "i'm a little pea")[0]
    assert user.groups == [foo_group]

    with testclient.session_transaction() as sess:
        del sess["sessions"]

    res = testclient.get(url, status=302)
    backend.delete(user)


def test_invitation_editable_user_name(
    testclient, logged_admin, foo_group, smtpd, backend
):
    assert backend.get(models.User, user_name="jackyjack") is None
    assert backend.get(models.User, user_name="djorje") is None

    res = testclient.get("/invite", status=200)

    res.form["user_name"] = "jackyjack"
    res.form["user_name_editable"] = True
    res.form["email"] = "jackyjack@domain.test"
    res.form["groups"] = [foo_group.id]
    res = res.form.submit(name="action", value="send", status=200)
    assert len(smtpd.messages) == 1

    url = res.pyquery(".copy-text")[0].value

    # logout
    g.session = None
    with testclient.session_transaction() as sess:
        del sess["sessions"]

    res = testclient.get(url, status=200)

    assert "readonly" not in res.form["user_name"].attrs
    assert "readonly" in res.form["emails-0"].attrs
    assert "readonly" in res.form["groups"].attrs

    assert res.form["user_name"].value == "jackyjack"
    assert res.form["emails-0"].value == "jackyjack@domain.test"
    assert res.form["groups"].value == [foo_group.id]

    res.form["user_name"] = "djorje"
    res.form["password1"] = "i'm a little pea"
    res.form["password2"] = "i'm a little pea"
    res.form["given_name"] = "George"
    res.form["family_name"] = "Abitbol"

    res = res.form.submit(status=302)

    assert ("success", "Your account has been created successfully.") in res.flashes
    res = res.follow(status=200)

    user = backend.get(models.User, user_name="djorje")
    backend.reload(foo_group)
    assert backend.check_user_password(user, "i'm a little pea")[0]
    assert user.groups == [foo_group]

    with testclient.session_transaction() as sess:
        del sess["sessions"]
    backend.delete(user)


def test_generate_link(testclient, logged_admin, foo_group, smtpd, backend):
    assert backend.get(models.User, user_name="sometwo") is None

    res = testclient.get("/invite", status=200)

    res.form["user_name"] = "sometwo"
    res.form["email"] = "sometwo@domain.test"
    res.form["groups"] = [foo_group.id]
    res = res.form.submit(name="action", value="generate", status=200)
    assert len(smtpd.messages) == 0

    url = res.pyquery(".copy-text")[0].value

    # logout
    g.session = None
    with testclient.session_transaction() as sess:
        del sess["sessions"]

    res = testclient.get(url, status=200)

    assert "readonly" in res.form["user_name"].attrs
    assert "readonly" in res.form["emails-0"].attrs
    assert "readonly" in res.form["groups"].attrs

    assert res.form["user_name"].value == "sometwo"
    assert res.form["emails-0"].value == "sometwo@domain.test"
    assert res.form["groups"].value == [foo_group.id]

    res.form["password1"] = "i'm a little pea"
    res.form["password2"] = "i'm a little pea"
    res.form["given_name"] = "George"
    res.form["family_name"] = "Abitbol"

    res = res.form.submit(status=302)
    res = res.follow(status=200)

    user = backend.get(models.User, user_name="sometwo")
    backend.reload(foo_group)
    assert backend.check_user_password(user, "i'm a little pea")[0]
    assert user.groups == [foo_group]

    with testclient.session_transaction() as sess:
        del sess["sessions"]

    res = testclient.get(url, status=302)
    backend.delete(user)


def test_invitation_login_already_taken(testclient, logged_admin):
    res = testclient.get("/invite", status=200)

    res.form["user_name"] = logged_admin.user_name
    res.form["email"] = logged_admin.preferred_email
    res = res.form.submit(name="action", value="send", status=200)

    res.mustcontain("The user name 'admin' already exists")
    res.mustcontain("The email 'jane@doe.test' is already used")


def test_registration(testclient, foo_group):
    payload = RegistrationPayload(
        datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [foo_group.id],
    )
    b64 = payload.b64()
    hash = payload.build_hash()

    testclient.get(f"/register/{b64}/{hash}", status=200)


def test_registration_formcontrol(testclient):
    payload = RegistrationPayload(
        datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [],
    )
    b64 = payload.b64()
    hash = payload.build_hash()

    res = testclient.get(f"/register/{b64}/{hash}", status=200)
    assert "emails-1" not in res.form.fields

    res = res.form.submit(status=200, name="fieldlist_add", value="phone_numbers-0")
    assert "phone_numbers-1" in res.form.fields


def test_registration_invalid_hash(testclient, foo_group):
    now = datetime.datetime.now(datetime.timezone.utc).isoformat()
    payload = RegistrationPayload(
        now, "anything", False, "someone@mydomain.test", [foo_group.id]
    )
    b64 = payload.b64()

    testclient.get(f"/register/{b64}/invalid", status=302)


def test_registration_invalid_data(testclient, foo_group):
    payload = RegistrationPayload(
        datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [foo_group.id],
    )
    hash = payload.build_hash()

    testclient.get(f"/register/invalid/{hash}", status=302)


def test_registration_more_than_48_hours_after_invitation(testclient, foo_group):
    two_days_ago = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
        hours=48
    )
    payload = RegistrationPayload(
        two_days_ago.isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [foo_group.id],
    )
    hash = payload.build_hash()
    b64 = payload.b64()

    testclient.get(f"/register/{b64}/{hash}", status=302)


def test_registration_no_password(testclient, foo_group, backend):
    payload = RegistrationPayload(
        datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [foo_group.id],
    )
    hash = payload.build_hash()
    b64 = payload.b64()
    url = f"/register/{b64}/{hash}"

    res = testclient.get(url, status=200)
    assert "required" in res.form["password1"].attrs
    assert "required" in res.form["password2"].attrs

    res = res.form.submit(status=200)
    res.mustcontain("This field is required.")

    assert not backend.get(models.User, user_name="someoneelse")

    with testclient.session_transaction() as sess:
        assert "sessions" not in sess


def test_no_registration_if_logged_in(testclient, logged_user, foo_group):
    payload = RegistrationPayload(
        datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [foo_group.id],
    )
    hash = payload.build_hash()
    b64 = payload.b64()
    url = f"/register/{b64}/{hash}"

    testclient.get(url, status=302)


def test_unavailable_if_no_smtp(testclient, logged_admin):
    res = testclient.get("/users")
    res.mustcontain("Invite")
    res = testclient.get("/profile")
    res.mustcontain("Invite")
    testclient.get("/invite")

    testclient.app.config["CANAILLE"]["SMTP"] = None

    res = testclient.get("/users")
    res.mustcontain(no="Invite")
    res = testclient.get("/profile")
    res.mustcontain(no="Invite")
    testclient.get("/invite", status=500, expect_errors=True)


def test_groups_are_saved_even_when_user_does_not_have_read_permission(
    testclient, foo_group, backend
):
    testclient.app.config["CANAILLE"]["ACL"]["DEFAULT"]["READ"] = [
        "user_name"
    ]  # remove groups from default read permissions

    payload = RegistrationPayload(
        datetime.datetime.now(datetime.timezone.utc).isoformat(),
        "someoneelse",
        False,
        "someone@mydomain.test",
        [foo_group.id],
    )
    b64 = payload.b64()
    hash = payload.build_hash()

    res = testclient.get(f"/register/{b64}/{hash}", status=200)

    assert res.form["groups"].value == [foo_group.id]
    assert "readonly" in res.form["groups"].attrs

    res.form["password1"] = "i'm a little pea"
    res.form["password2"] = "i'm a little pea"
    res.form["given_name"] = "George"
    res.form["family_name"] = "Abitbol"

    res = res.form.submit(status=302)
    res = res.follow(status=200)

    user = backend.get(models.User, user_name="someoneelse")
    backend.reload(foo_group)
    assert user.groups == [foo_group]
    backend.delete(user)



import json

import wtforms
from joserfc.jwk import KeySet

from canaille.app import models
from canaille.app.forms import Form
from canaille.app.forms import IDToModel
from canaille.app.forms import email_validator
from canaille.app.forms import is_uri
from canaille.app.forms import unique_values
from canaille.app.i18n import lazy_gettext as _
from canaille.backends import Backend


class AuthorizeForm(Form):
    pass


class LogoutForm(Form):
    answer = wtforms.SubmitField()


def is_jwks(form, field):
    try:
        payload = json.loads(field.data)
    except json.decoder.JSONDecodeError as exc:
        raise wtforms.ValidationError(
            _("This value is not a valid JSON string.")
        ) from exc

    try:
        KeySet.import_key_set(payload)
    except (KeyError, ValueError) as exc:
        raise wtforms.ValidationError(_("This value is not a valid JWK.")) from exc


def _client_audiences():
    return [
        (client, client.client_name) for client in Backend.instance.query(models.Client)
    ]


class ClientAddForm(Form):
    client_name = wtforms.StringField(
        _("Name"),
        validators=[wtforms.validators.DataRequired()],
        render_kw={"placeholder": "Client Name"},
        description=_(
            "This is the name that will be displayed on the web interface, and notably on the consent page."
        ),
    )

    contacts = wtforms.FieldList(
        wtforms.EmailField(
            _("Contacts"),
            validators=[wtforms.validators.Optional(), email_validator],
            render_kw={"placeholder": "admin@client.example"},
        ),
        min_entries=1,
        validators=[unique_values],
        description=_(
            "Those are the email addresses of people responsible for the client."
        ),
    )

    client_uri = wtforms.URLField(
        _("URI"),
        validators=[
            wtforms.validators.DataRequired(),
            is_uri,
        ],
        render_kw={"placeholder": "https://client.example"},
        description=_(
            "URL string of a web page providing information about the client."
        ),
    )

    redirect_uris = wtforms.FieldList(
        wtforms.URLField(
            _("Redirect URIs"),
            validators=[
                wtforms.validators.DataRequired(),
                is_uri,
            ],
            render_kw={"placeholder": "https://client.example/oauth/callback"},
        ),
        min_entries=1,
        validators=[unique_values],
        description=_(
            "URIs for use in redirect-based flows such as the authorization code and implicit flows."
        ),
    )
    post_logout_redirect_uris = wtforms.FieldList(
        wtforms.URLField(
            _("Post logout redirect URIs"),
            validators=[
                wtforms.validators.Optional(),
                is_uri,
            ],
            render_kw={"placeholder": "https://client.example/oauth/logout-callback"},
        ),
        min_entries=1,
        validators=[unique_values],
        description=_(
            "URIs that the client may redirect users to after logging them out."
        ),
    )

    grant_types = wtforms.SelectMultipleField(
        _("Grant types"),
        validators=[wtforms.validators.DataRequired()],
        choices=[
            ("password", "password"),
            ("authorization_code", "authorization_code"),
            ("implicit", "implicit"),
            ("hybrid", "hybrid"),
            ("refresh_token", "refresh_token"),
            ("client_credentials", "client_credentials"),
            ("urn:ietf:params:oauth:grant-type:jwt-bearer", "jwt-bearer"),
        ],
        default=["authorization_code", "refresh_token"],
        description=_("Grant types that the client can use at the token endpoint."),
    )

    scope = wtforms.StringField(
        _("Scope"),
        validators=[wtforms.validators.DataRequired()],
        default="openid profile email",
        render_kw={"placeholder": "openid profile"},
        description=_("Kind of information that the client can request about users."),
    )

    response_types = wtforms.SelectMultipleField(
        _("Response types"),
        validators=[wtforms.validators.DataRequired()],
        choices=[("code", "code"), ("token", "token"), ("id_token", "id_token")],
        default=["code"],
        description=_(
            "Response types that the client can use at the authorization endpoint."
        ),
    )

    token_endpoint_auth_method = wtforms.SelectField(
        _("Token endpoint authentication method"),
        validators=[wtforms.validators.DataRequired()],
        choices=[
            ("client_secret_basic", "client_secret_basic"),
            ("client_secret_post", "client_secret_post"),
            ("client_secret_jwt", "client_secret_jwt"),
            ("private_key_jwt", "private_key_jwt"),
            ("none", "none"),
        ],
        default="client_secret_basic",
        description=_(
            "Authentication method that the client will use at the token endpoint."
        ),
    )

    audience = wtforms.SelectMultipleField(
        _("Token audiences"),
        validators=[wtforms.validators.Optional()],
        choices=_client_audiences,
        validate_choice=False,
        coerce=IDToModel("Client"),
        description=_(
            "The other clients that are intended to use the tokens emitted by this client."
        ),
    )

    logo_uri = wtforms.URLField(
        _("Logo URI"),
        validators=[
            wtforms.validators.Optional(),
            is_uri,
        ],
        render_kw={"placeholder": "https://client.example/logo.png"},
        description=_("An URL for the logo of this client."),
    )

    tos_uri = wtforms.URLField(
        _("Terms of service URI"),
        validators=[
            wtforms.validators.Optional(),
            is_uri,
        ],
        render_kw={"placeholder": "https://client.example/tos.html"},
        description=_("URL to the terms of service of the client."),
    )

    policy_uri = wtforms.URLField(
        _("Policy URI"),
        validators=[
            wtforms.validators.Optional(),
            is_uri,
        ],
        render_kw={"placeholder": "https://client.example/policy.html"},
        description=_("URL to the privacy policy of the client."),
    )

    software_id = wtforms.StringField(
        _("Software ID"),
        validators=[wtforms.validators.Optional()],
        render_kw={"placeholder": "xyz"},
        description=_(
            "Unique identifier for this client, that should be stable in time and common for all identity providers."
        ),
    )

    software_version = wtforms.StringField(
        _("Software Version"),
        validators=[wtforms.validators.Optional()],
        render_kw={"placeholder": "1.0"},
        description=_("The version of the client."),
    )

    jwks = wtforms.StringField(
        _("JSON Web Keys"),
        validators=[wtforms.validators.Optional(), is_jwks],
        render_kw={"placeholder": ""},
        description=_("The public JSON Web Keys of the client."),
    )

    jwks_uri = wtforms.URLField(
        _("JSON Web Keys URI"),
        validators=[
            wtforms.validators.Optional(),
            is_uri,
        ],
        render_kw={"placeholder": "https://client.example/oauth/jwks.json"},
        description=_("The URI that points to the public JSON Web Keys of the client."),
    )


class TokenRevokationForm(Form):
    pass



import datetime

from scim2_client.engines.werkzeug import TestSCIMClient
from scim2_models import Error
from scim2_models import Resource
from werkzeug.security import gen_salt
from werkzeug.test import Client

from canaille.app import models
from canaille.scim.endpoints import bp
from canaille.scim.endpoints import get_resource_types
from canaille.scim.endpoints import get_schemas
from canaille.scim.endpoints import get_service_provider_config


def test_authentication_failure(app):
    """Test authentication with an invalid token."""
    resource_models = [
        Resource.from_schema(schema) for schema in get_schemas().values()
    ]
    scim_client = TestSCIMClient(
        Client(app),
        scim_prefix=bp.url_prefix,
        environ={"headers": {"Authorization": "Bearer invalid"}},
        service_provider_config=get_service_provider_config(),
        resource_types=get_resource_types().values(),
        resource_models=resource_models,
    )
    User = scim_client.get_resource_model("User")
    error = scim_client.query(User, raise_scim_errors=False)
    assert isinstance(error, Error)
    assert not error.scim_type
    assert error.status == 401


def test_authentication_with_an_user_token(app, backend, oidc_client, user):
    """Test authentication with an user token."""
    scim_token = models.Token(
        token_id=gen_salt(48),
        access_token=gen_salt(48),
        subject=user,
        audience=[oidc_client],
        client=oidc_client,
        refresh_token=gen_salt(48),
        scope=["openid", "profile"],
        issue_date=datetime.datetime.now(datetime.timezone.utc),
        lifetime=3600,
    )
    backend.save(scim_token)

    resource_models = [
        Resource.from_schema(schema) for schema in get_schemas().values()
    ]
    scim_client = TestSCIMClient(
        Client(app),
        scim_prefix=bp.url_prefix,
        environ={"headers": {"Authorization": f"Bearer {scim_token.access_token}"}},
        service_provider_config=get_service_provider_config(),
        resource_types=get_resource_types().values(),
        resource_models=resource_models,
    )
    User = scim_client.get_resource_model("User")
    error = scim_client.query(User, raise_scim_errors=False)
    assert isinstance(error, Error)
    assert not error.scim_type
    assert error.status == 401


def test_missing_field(app, backend, scim_client, scim_token):
    scim_client.discover()
    error = scim_client.client.post(
        "/scim/v2/Users",
        json={"foo": "bar"},
        headers={"Authorization": f"Bearer {scim_token.access_token}"},
    )
    assert error.json == {
        "detail": "Extra inputs are not permitted: foo",
        "schemas": ["urn:ietf:params:scim:api:messages:2.0:Error"],
        "status": "400",
    }


def test_invalid_payload(app, backend, scim_client):
    # TODO: push this test in scim2-tester
    scim_client.discover()
    User = scim_client.get_resource_model("User")
    error = scim_client.create(User(), raise_scim_errors=False)
    assert isinstance(error, Error)
    assert error.scim_type == "invalidValue"
    assert error.status == 400
